#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains widget with geometry functionality
"""

from __future__ import print_function, division, absolute_import

import logging

from Qt.QtCore import QObject

from tpDcc.libs.qt.core import qtutils

from tpRigToolkit.tools.rigtoolbox.widgets import base

LOGGER = logging.getLogger('tpRigToolkit-tools-rigtoolbox')


class GeneralWidget(base.CommandRigToolBoxWidget, object):

    def __init__(self, client, commands_data, parent=None):

        self._model = GeneralWidgetModel()

        super(GeneralWidget, self).__init__(
            title='General', commands_data=commands_data,
            controller=GeneralWidgetController(model=self._model, client=client), parent=parent)


class GeneralWidgetModel(QObject, object):
    def __init__(self):
        super(GeneralWidgetModel, self).__init__()


class GeneralWidgetController(object):

    def __init__(self, model, client):
        super(GeneralWidgetController, self).__init__()

        self._model = model
        self._client = client

    @property
    def model(self):
        return self._model

    @property
    def client(self):
        return self._client

    def delete_history(self):
        return self._client.delete_history()

    def freeze_transforms(self):
        return self._client.freeze_transforms()

    def move_pivot_to_zero(self):
        return self._client.move_pivot_to_zero()

    def lock_all_transforms(self):
        return self._client.lock_all_transforms()

    def lock_translation(self):
        return self._client.lock_translation()

    def lock_rotation(self):
        return self._client.lock_rotation()

    def lock_scale(self):
        return self._client.lock_scale()

    def lock_visibility(self):
        return self._client.lock_visibility()

    def unlock_all_transforms(self):
        return self._client.unlock_all_transforms()

    def unlock_translation(self):
        return self._client.unlock_translation()

    def unlock_rotation(self):
        return self._client.unlock_rotation()

    def unlock_scale(self):
        return self._client.unlock_scale()

    def unlock_visibility(self):
        return self._client.unlock_visibility()

    def clean_student_license(self):
        return self._client.clean_student_license()

    def match_transform(self):
        return self._client.match_transform()

    def match_translation(self):
        return self._client.match_translation()

    def match_rotation(self):
        return self._client.match_rotation()

    def match_scale(self):
        return self._client.match_scale()

    def combine_meshes(self):
        new_mesh_name = None
        if qtutils.is_shift_modifier():
            new_mesh_name = qtutils.get_string_input('Combine Mesh Name', title='Combine Mesh')

        return self._client.combine_meshes(new_mesh_name=new_mesh_name)

    def separate_meshes(self):
        return self._client.separate_meshes()

    def mirror_meshes(self):
        return self._client.mirror_meshes()

    def mirror_meshes_options(self):
        return self._client.mirror_meshes_options()

    def open_symmetry_tool(self):
        return self._client.open_symmetry_tool()

    def detach_components(self):
        return self._client.detach_components()

    def curve_from_edge_loop(self):
        return self._client.curve_from_edge_loop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Collection of rigging tools
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import logging

from tpDcc.core import tool
from tpDcc.managers import resources

from tpRigToolkit.tools.rigtoolbox.core import consts, client, toolset

LOGGER = logging.getLogger(consts.TOOL_ID)


class RigToolboxTool(tool.DccTool, object):

    ID = consts.TOOL_ID
    CLIENT_CLASS = client.RigToolboxClient
    TOOLSET_CLASS = toolset.RigToolboxToolset

    def __init__(self, *args, **kwargs):
        super(RigToolboxTool, self).__init__(*args, **kwargs)

        resources_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'resources')
        resources.register_resource(resources_path)

    @classmethod
    def config_dict(cls, file_name=None):
        base_tool_config = tool.DccTool.config_dict(file_name=file_name)
        tool_config = {
            'name': 'Rig ToolBox',
            'id': RigToolboxTool.ID,
            'supported_dccs': {'maya': ['2017', '2018', '2019', '2020']},
            'logo': 'rigtoolbox',
            'icon': 'rigtoolbox',
            'tooltip': 'Collection of tools to improve the rigging process',
            'tags': ['tpRigToolkit', 'rig', 'tools', 'toolbox'],
            'logger_dir': os.path.join(os.path.expanduser('~'), 'tpRigToolkit', 'logs', 'tools'),
            'logger_level': 'INFO',
            'is_checkable': False,
            'is_checked': False,
            'menu_ui': {'label': 'Rig ToolBox', 'load_on_startup': False, 'color': '', 'background_color': ''}
        }
        base_tool_config.update(tool_config)

        return base_tool_config

    def launch(self, *args, **kwargs):
        return self.launch_frameless(*args, **kwargs)


if __name__ == '__main__':
    import tpRigToolkit.loader
    from tpDcc.managers import tools

    tool_path = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))))
    if tool_path not in sys.path:
        sys.path.append(tool_path)

    tpRigToolkit.loader.init()
    tools.ToolsManager().launch_tool_by_id(consts.TOOL_ID)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains dialog to assign labels to joints
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QObject
from Qt.QtWidgets import QDialogButtonBox, QMessageBox

from tpDcc.libs.qt.core import qtutils
from tpDcc.libs.qt.widgets import layouts, dialog, label, lineedit, dividers, message


class JointsLabelDialog(dialog.BaseDialog, object):
    def __init__(self, parent=None):

        self._model = JointsLabelModel()
        self._controller = JointsLabelController(model=self._model)

        super(JointsLabelDialog, self).__init__(
            title='Joints Label Dialog', size=(200, 50), frameless=True, parent=parent
        )

        self._dragger._button_closed.setVisible(False)
        self.refresh()
        self.setFixedHeight(250)
        self.setFixedWidth(300)

    def ui(self):
        super(JointsLabelDialog, self).ui()

        name_label = label.BaseLabel('Specify label sides', parent=self)
        left_side_label = label.BaseLabel('Left Side: ', parent=self)
        self._left_side_line = lineedit.BaseLineEdit(parent=self)
        right_side_label = label.BaseLabel('Right Side: ', parent=self)
        self._right_side_line = lineedit.BaseLineEdit(parent=self)

        info_widget = message.BaseMessage(
            'You can disable auto joint label in tool\n button options (right click on it', parent=self).info()

        grid_layout = layouts.GridLayout()
        grid_layout.addWidget(name_label, 0, 0, 1, 2)
        grid_layout.addWidget(left_side_label, 1, 0, Qt.AlignRight)
        grid_layout.addWidget(self._left_side_line, 1, 1)
        grid_layout.addWidget(right_side_label, 2, 0, Qt.AlignRight)
        grid_layout.addWidget(self._right_side_line, 2, 1)

        self._button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        self.main_layout.addLayout(grid_layout)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addWidget(self._button_box)
        self.main_layout.addWidget(dividers.Divider())
        self.main_layout.addWidget(info_widget)

    def setup_signals(self):
        self._left_side_line.textChanged.connect(self._controller.set_left_side)
        self._right_side_line.textChanged.connect(self._controller.set_right_side)
        self._button_box.accepted.connect(self._on_accept)
        self._button_box.rejected.connect(self._on_reject)

        self._model.leftSideChanged.connect(self._left_side_line.setText)
        self._model.rightSideChanged.connect(self._right_side_line.setText)

    def refresh(self):
        self._left_side_line.setText(self._model.left_side)
        self._right_side_line.setText(self._model.right_side)

    def get_sides(self):
        return self._model.left_side, self._model.right_side

    def _on_accept(self):
        valid_verify = self._controller.verify()
        if valid_verify:
            self.accept()
            return True

        res = qtutils.show_warning(
            self, 'Missing side data', 'Please fill both side fields. Do you want to continue with default settings?')
        if res != QMessageBox.Yes:
            return True

        self.reject()

    def _on_reject(self):
        self._controller.set_left_side('')
        self._controller.set_right_side('')
        self.reject()


class JointsLabelModel(QObject, object):

    leftSideChanged = Signal(str)
    rightSideChanged = Signal(str)

    def __init__(self):
        super(JointsLabelModel, self).__init__()

        self._left_side = '*_l_*'
        self._right_side = '*_r_*'

    @property
    def left_side(self):
        return self._left_side

    @left_side.setter
    def left_side(self, value):
        self._left_side = str(value)
        self.leftSideChanged.emit(self._left_side)

    @property
    def right_side(self):
        return self._right_side

    @right_side.setter
    def right_side(self, value):
        self._right_side = str(value)
        self.rightSideChanged.emit(self._right_side)


class JointsLabelController(object):
    def __init__(self, model):
        self._model = model

    @property
    def model(self):
        return self._model

    def set_left_side(self, value):
        self._model.left_side = value

    def set_right_side(self, value):
        self._model.right_side = value

    def verify(self):
        left_side = self._model.left_side
        right_side = self._model.right_side

        if left_side and right_side:
            return True

        return False



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library with general functions
"""

from __future__ import print_function, division, absolute_import

import re

import maya.cmds
import maya.mel

from tpDcc import dcc
from tpDcc.managers import tools
from tpDcc.libs.python import python
from tpDcc.dccs.maya.core import decorators, transform as xform_utils, curve as curve_utils

from tpRigToolkit.tools.rigtoolbox.widgets import library


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def delete_history(transforms=None):
    """
    Delete history of selected transforms
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No nodes to delete history of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Deleting history: {}'.format(node))
        try:
            dcc.delete_history(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to delete history in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def freeze_transforms(transforms=None):
    """
    Freeze selected transforms
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to freeze transforms of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Freezing transforms: {}'.format(node))
        try:
            dcc.freeze_transforms(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to freeze transforms in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def move_pivot_to_zero(transforms=None):
    """
    Moves selected nodes pivots to zero (0, 0, 0 in the world)
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to move pivot to zero of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Moving pivot to zero: {}'.format(node))
        try:
            dcc.move_pivot_to_zero(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible move pivot to zero for node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def lock_all_transforms(transforms=None):
    """
    Locks all the transform channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to lock all transform channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Locking all transforms: {}'.format(node))
        try:
            dcc.lock_translate_attributes(node)
            dcc.lock_rotate_attributes(node)
            dcc.lock_scale_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to lock all transforms for node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def lock_translation(transforms=None):
    """
    Locks all translation channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to lock all translation channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Locking translation channels: {}'.format(node))
        try:
            dcc.lock_translate_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to lock translate channels in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def lock_rotation(transforms=None):
    """
    Locks all rotation channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to lock all rotation channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Locking rotation channels: {}'.format(node))
        try:
            dcc.lock_rotate_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to lock rotations channel in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def lock_scale(transforms=None):
    """
    Locks all scale channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to lock all scale channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Locking scale channels: {}'.format(node))
        try:
            dcc.lock_scale_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to lock scale channels in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def lock_visibility(transforms=None):
    """
    Locks visibility channel of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to lock all scale channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Locking visibility channel: {}'.format(node))
        try:
            dcc.lock_visibility_attribute(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to lock visibility channel in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def unlock_all_transforms(transforms=None):
    """
    Locks all the transform channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to unlock all transform channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Unlocking all transforms: {}'.format(node))
        try:
            dcc.unlock_translate_attributes(node)
            dcc.unlock_rotate_attributes(node)
            dcc.unlock_scale_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to unlock all transforms for node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def unlock_translation(transforms=None):
    """
    Locks all translation channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to unlock all translation channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Unlocking translation channels: {}'.format(node))
        try:
            dcc.unlock_translate_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to unlock translate channels in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def unlock_rotation(transforms=None):
    """
    Locks all rotation channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to unlock all rotation channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Unlocking rotation channels: {}'.format(node))
        try:
            dcc.unlock_rotate_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to unlock rotations channel in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def unlock_scale(transforms=None):
    """
    Locks all scale channels of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to unlock all scale channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Unlocking scale channels: {}'.format(node))
        try:
            dcc.unlock_scale_attributes(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to unlock scale channels in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def unlock_visibility(transforms=None):
    """
    Locks visibility channel of the given transforms nodes
    """

    out_dict = {'success': False, 'result': list()}

    transforms = python.force_list(transforms or dcc.selected_nodes_of_type(node_type='transform') or list())
    if not transforms:
        out_dict['msg'] = 'No transforms to unlock all scale channels of. Select at least one.'
        return out_dict

    percentage = 100.0 / len(transforms)

    for i, node in enumerate(transforms):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Unlocking scale channels: {}'.format(node))
        try:
            dcc.unlock_visibility_attribute(node)
            out_dict['result'].append(node)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to unlock visibility channel in node: "{}" : {}'.format(node, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def match_transform(source_transform=None, target_transform=None):
    """
    Matches all the transforms of the source node to the transforms of the given target node(s)
    """

    out_dict = {'success': False, 'result': list()}

    selection = dcc.selected_nodes_of_type(node_type='transform')
    source_transform = source_transform or selection[0] if python.index_exists_in_list(selection, 0) else None
    if not source_transform:
        out_dict['msg'] = 'No source transform given to match against target transform.'
        return out_dict
    target_transform = target_transform or selection[1:] if len(selection) > 1 else None
    if not source_transform:
        out_dict['msg'] = 'No target transform(s) given to match source transform against.'
        return out_dict
    source_transform = python.force_list(source_transform)
    target_transform = python.force_list(target_transform)

    percentage = 100.0 / len(source_transform)

    for i, source in enumerate(source_transform):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Matching transforms: {}'.format(source))
        try:
            maya.cmds.delete(maya.cmds.parentConstraint(target_transform, source, maintainOffset=False))
            maya.cmds.delete(maya.cmds.scaleConstraint(target_transform, source, maintainOffset=False))

            # For joints, we store now rotation data in jointOrient attribute
            if dcc.node_type(source) == 'joint':
                for axis in 'XYZ':
                    joint_orient_attr = 'jointOrient{}'.format(axis)
                    joint_rotation_attr = 'rotate{}'.format(axis)
                    dcc.set_attribute_value(source, joint_orient_attr, 0.0)
                    joint_rotation = dcc.get_attribute_value(source, joint_rotation_attr)
                    dcc.set_attribute_value(source, joint_orient_attr, joint_rotation)
                    dcc.set_attribute_value(source, joint_rotation_attr, 0.0)

            out_dict['result'].append(source)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to match node "{}" transforms to "{}" : {}'.format(
                source_transform, target_transform, exc)
            return out_dict

    matched_nodes = out_dict.get('result', None)
    if matched_nodes:
        dcc.select_node(matched_nodes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def match_translation(source_transform=None, target_transform=None):
    """
    Matches translation of the source node to the translation of the given target node(s)
    """

    out_dict = {'success': False, 'result': list()}

    selection = dcc.selected_nodes_of_type(node_type='transform')
    source_transform = source_transform or selection[0] if python.index_exists_in_list(selection, 0) else None
    if not source_transform:
        out_dict['msg'] = 'No source transform given to match against target translation.'
        return out_dict
    target_transform = target_transform or selection[1:] if len(selection) > 1 else None
    if not source_transform:
        out_dict['msg'] = 'No target transform(s) given to match source translation against.'
        return out_dict
    source_transform = python.force_list(source_transform)
    target_transform = python.force_list(target_transform)

    percentage = 100.0 / len(source_transform)

    for i, source in enumerate(source_transform):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Matching translation: {}'.format(source))
        try:
            maya.cmds.delete(maya.cmds.pointConstraint(target_transform, source, maintainOffset=False))
            out_dict['result'].append(source)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to match node "{}" translation to "{}" : {}'.format(
                source_transform, target_transform, exc)
            return out_dict

    matched_nodes = out_dict.get('result', None)
    if matched_nodes:
        dcc.select_node(matched_nodes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def match_rotation(source_transform=None, target_transform=None):
    """
    Matches rotation of the source node to the rotation of the given target node(s)
    """

    out_dict = {'success': False, 'result': list()}

    selection = dcc.selected_nodes_of_type(node_type='transform')
    source_transform = source_transform or selection[0] if python.index_exists_in_list(selection, 0) else None
    if not source_transform:
        out_dict['msg'] = 'No source transform given to match against target rotation.'
        return out_dict
    target_transform = target_transform or selection[1:] if len(selection) > 1 else None
    if not source_transform:
        out_dict['msg'] = 'No target transform(s) given to match source rotation against.'
        return out_dict
    source_transform = python.force_list(source_transform)
    target_transform = python.force_list(target_transform)

    percentage = 100.0 / len(source_transform)

    for i, source in enumerate(source_transform):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Matching rotation: {}'.format(source))
        try:
            maya.cmds.delete(maya.cmds.orientConstraint(target_transform, source, maintainOffset=False))

            # For joints, we store now rotation data in jointOrient attribute
            if dcc.node_type(source) == 'joint':
                for axis in 'XYZ':
                    joint_orient_attr = 'jointOrient{}'.format(axis)
                    joint_rotation_attr = 'rotate{}'.format(axis)
                    dcc.set_attribute_value(source, joint_orient_attr, 0.0)
                    joint_rotation = dcc.get_attribute_value(source, joint_rotation_attr)
                    dcc.set_attribute_value(source, joint_orient_attr, joint_rotation)
                    dcc.set_attribute_value(source, joint_rotation_attr, 0.0)

            out_dict['result'].append(source)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to match node "{}" rotation to "{}" : {}'.format(
                source_transform, target_transform, exc)
            return out_dict

    matched_nodes = out_dict.get('result', None)
    if matched_nodes:
        dcc.select_node(matched_nodes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def match_scale(source_transform=None, target_transform=None):
    """
    Matches scale of the source node to the scale of the given target node(s)
    """

    out_dict = {'success': False, 'result': list()}

    selection = dcc.selected_nodes_of_type(node_type='transform')
    source_transform = source_transform or selection[0] if python.index_exists_in_list(selection, 0) else None
    if not source_transform:
        out_dict['msg'] = 'No source transform given to match against target scale.'
        return out_dict
    target_transform = target_transform or selection[1:] if len(selection) > 1 else None
    if not source_transform:
        out_dict['msg'] = 'No target transform(s) given to match source scale against.'
        return out_dict
    source_transform = python.force_list(source_transform)
    target_transform = python.force_list(target_transform)

    percentage = 100.0 / len(source_transform)

    for i, source in enumerate(source_transform):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Matching scale: {}'.format(source))
        try:
            maya.cmds.delete(maya.cmds.scaleConstraint(target_transform, source, maintainOffset=False))
            out_dict['result'].append(source)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to match node "{}" scale to "{}" : {}'.format(
                source_transform, target_transform, exc)
            return out_dict

    matched_nodes = out_dict.get('result', None)
    if matched_nodes:
        dcc.select_node(matched_nodes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def combine_meshes(meshes=None, new_mesh_name=None):
    """
    Combines given meshes into one transform
    """

    out_dict = {'success': False, 'result': None}

    meshes = meshes or dcc.selected_nodes_of_type(node_type='transform')
    meshes = python.force_list(meshes)
    if not meshes:
        out_dict['msg'] = 'No meshes to combine selected.'
        return out_dict
    if len(meshes) < 2:
        out_dict['msg'] = 'You need to select at least two meshes to combine.'
        return out_dict
    new_mesh_name = new_mesh_name or dcc.node_short_name(meshes[0])

    parent_node = None
    node_parents = list(set([dcc.node_parent(mesh) for mesh in meshes]))
    if all(parent == node_parents[0] for parent in node_parents):
        parent_node = node_parents[0]

    try:
        combined_mesh = dcc.combine_meshes(construction_history=False)
        if not combined_mesh:
            out_dict['msg'] = 'Combine operation was done but not combined mesh was generated'
            return out_dict
        combined_mesh = dcc.rename_node(combined_mesh, new_mesh_name)
        if parent_node:
            dcc.set_parent(combined_mesh, parent_node)

        out_dict['result'] = combined_mesh
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to combine meshes "{}" : {}'.format(meshes, exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def separate_meshes(meshes=None, new_mesh_name=None):
    """
    Separates given meshes into one transform
    """

    out_dict = {'success': False, 'result': None}

    meshes = meshes or dcc.selected_nodes_of_type(node_type='transform')
    meshes = python.force_list(meshes)
    if not meshes:
        out_dict['msg'] = 'No meshes to separate selected.'
        return out_dict
    new_mesh_name = new_mesh_name or dcc.node_short_name(meshes[0])

    try:
        result_meshes = list()
        separated_meshes = dcc.separate_meshes(construction_history=False)
        if not separated_meshes:
            out_dict['msg'] = 'Separate operation was done but not separated mesh was generated'
            return out_dict
        for separated_mesh in separated_meshes:
            separated_mesh = dcc.rename_node(separated_mesh, new_mesh_name)
            result_meshes.append(separated_mesh)

        out_dict['result'] = result_meshes
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to separate meshes "{}" : {}'.format(meshes, exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def mirror_mesh(mesh=None):
    """
    Mirror given meshes
    """

    out_dict = {'success': False, 'result': list()}

    meshes = mesh or dcc.selected_nodes_of_type(node_type='transform')
    meshes = python.force_list(meshes)
    if not meshes:
        out_dict['msg'] = 'No meshes to mirror selected.'
        return out_dict

    for mesh in meshes:
        try:
            parent_node = dcc.node_parent(mesh)
            mirror_geo_name = xform_utils.find_transform_right_side(mesh, check_if_exists=False)
            mirror_geo = maya.cmds.duplicate(mesh, n=mirror_geo_name or None)
            root = maya.cmds.group(empty=True, world=True)
            maya.cmds.parent(mirror_geo, root)
            maya.cmds.setAttr('{}.rx'.format(root), 180)
            for axis in 'xyz':
                maya.cmds.setAttr('{}.s{}'.format(root, axis), -1)
            maya.cmds.parent(mirror_geo, world=True)
            maya.cmds.delete(root)
            maya.cmds.makeIdentity(mirror_geo, apply=True, t=False, r=True, s=True, n=False, pn=True)
            if parent_node:
                dcc.set_parent(mirror_geo, parent_node)
            out_dict['result'].append(mirror_geo)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to mirror meshes "{}" : {}'.format(meshes, exc)
            return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def open_mirror_meshes_options():

    out_dict = {'success': False}

    try:
        maya.mel.eval('MirrorPolygonGeometryOptions')
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to open mirror meshes options : {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


def open_symmetry_tool():

    out_dict = {'success': False, 'result': None}

    try:
        tool = tools.ToolsManager().launch_tool_by_id('tpRigToolkit-tools-symmesh')
        out_dict['result'] = tool
    except Exception as exc:
        out_dict['msg'] = 'Was not to open symmetry tool : {} '.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def detach_components(components=None):
    """
    Detach selected components
    """

    out_dict = {'success': False, 'result': list()}

    selection = components or dcc.selected_nodes()
    if not selection:
        out_dict['msg'] = 'No components to detach selected.'
        return out_dict

    try:
        maya.cmds.DetachComponent()
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to detach components "{}" : '.format(selection, exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def detach_edges(edges=None):
    """
    Detach selected edges in different groups
    """

    # make list of edges, that belong to certain object
    def _group_edges(obj, edges_list):
        return [item for item in edges_list if re.match(obj, item)]

    out_dict = {'success': False, 'result': list()}

    valid_edges = list()
    selection = edges or dcc.selected_nodes()
    selection = python.force_list(selection)
    for obj in selection:
        obj_type = maya.cmds.nodeType(obj)
        if obj_type == 'transform':
            continue
        is_edge = maya.cmds.filterExpand(sm=32)
        if is_edge is None:
            continue
        valid_edges.append(obj)

    if not valid_edges:
        out_dict['msg'] = 'No edges to detach selected.'
        return out_dict

    obj_list = [edge.split('.')[0] for edge in valid_edges]
    edge_list = [_group_edges(obj, valid_edges) for obj in obj_list]

    try:
        for edge_group in edge_list:
            maya.cmds.select(edge_group)
            maya.cmds.polySplitEdge()
            out_dict['result'].append(edge_group)
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to detach edges "{}" : '.format(valid_edges, exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def create_curve_from_mesh_edge_loop(mesh_edge_list=None):
    """
    Creates new curve from given mesh edge loop
    """

    out_dict = {'success': False, 'result': list()}

    if not mesh_edge_list:
        mesh_edge_list = maya.cmds.ls(sl=True, flatten=True)
    if not mesh_edge_list:
        out_dict['msg'] = 'No mesh edge loop to create curve from.'
        return out_dict

    try:
        new_curve = curve_utils.create_curve_from_mesh_edge_loop(mesh_edge_list=mesh_edge_list)
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to create curve from list of edges: {}'.format(exc)
        return out_dict

    out_dict['result'] = new_curve
    out_dict['success'] = True

    return out_dict



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains fallof curve widget implementation
"""

from Qt.QtCore import Qt, Signal, QObject, QPoint, QPointF, QRect, QRectF, QLineF
from Qt.QtWidgets import QApplication, QMenuBar, QUndoStack, QUndoCommand, QGraphicsView
from Qt.QtWidgets import QGraphicsItem, QGraphicsPathItem, QGraphicsScene, QGraphicsLineItem, QGraphicsRectItem
from Qt.QtGui import QColor, QPainterPath, QPainter, QPen, QBrush, QRadialGradient, QKeySequence, QKeyEvent

from tpDcc.libs.python import bezier
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, dividers, combobox, checkbox


@theme.mixin
class FallofCurveWidget(base.BaseWidget, object):

    curveUpdated = Signal(list)

    def __init__(self, parent=None):
        self._base_size = 300
        self._bezier_dict = {
            'bezier': [
                QPoint(0, 0), QPoint(75, 0),
                QPoint(225, 300), QPoint(300, 300)],
            'linear': [
                QPointF(0.000000, 0.000000), QPointF(75.000000, 75.000000),
                QPointF(225.000000, 225.000000), QPointF(300.000000, 300.000000)]
        }

        super(FallofCurveWidget, self).__init__(parent=parent)
        self.setObjectName('Falloff Curve')
        self.show()

        self.refresh()

    def ui(self):
        super(FallofCurveWidget, self).ui()

        base_rect = QRect(0, 0, self._base_size, self._base_size)

        self._scene = CurveNodeScene(base_rect)
        self._scene.base_size = self._base_size
        self._view = CurveNodeView(parent=self)
        self._view.setScene(self._scene)
        self._view.setGeometry(base_rect)

        self._menu_bar = QMenuBar(self)
        self._menu_bar.addAction(self._scene.undo_action)
        self._menu_bar.addAction(self._scene.redo_action)

        bottom_layout = layouts.HorizontalLayout(spacing=2, margins=(2, 2, 2, 2))
        self._bezier_type_combo = combobox.BaseComboBox(parent=self)
        self._snap_cbx = checkbox.BaseCheckBox('Snap', parent=self)
        bottom_layout.addWidget(self._bezier_type_combo)
        bottom_layout.addWidget(self._snap_cbx)

        self.main_layout.addWidget(self._menu_bar)
        self.main_layout.addWidget(dividers.Divider(parent=self))
        self.main_layout.addWidget(self._view)
        self.main_layout.addWidget(dividers.Divider(parent=self))
        self.main_layout.addLayout(bottom_layout)

    def setup_signals(self):
        self._bezier_type_combo.currentTextChanged.connect(self._on_change_curve)
        self._snap_cbx.toggled.connect(self._on_change_snap)
        self._scene.curveUpdated.connect(self._on_curve_updated)

    def resizeEvent(self, event):
        super(FallofCurveWidget, self).resizeEvent(event)
        self.update_view()

    def update_view(self):
        self._view.fitInView(self._scene.itemsBoundingRect(), Qt.KeepAspectRatio)

    def refresh(self):
        self.update_view()
        self._bezier_type_combo.clear()
        for key, value in self._bezier_dict.items():
            self._bezier_type_combo.addItem(key)
        self._scene.undo_stack.clear()

    def curve_as_points(self):
        points = self._scene.bezier_curve.points
        points_list = list()
        for point in points:
            points_list.append([point.x(), point.y()])

        return points_list

    def get_division_data(self, divisions=11):
        percentage = 1 / (divisions - 1.0)
        divisions_list = list()
        for i in range(divisions):
            divisions_list.append(i * percentage)

        return self.get_data_on_points(divisions_list)

    def get_data_on_percentage(self, percentage, points_list=None):
        points_list = points_list or self.curve_as_points()
        return bezier.get_data_on_percentage(percentage, points_list)

    def get_data_on_points(self, in_list=None):
        in_list = in_list or [0.0, 0.2, 0.25, 0.5, 0.6, 0.66, 0.8, 1.0]
        percentage_list = list()
        for point in in_list:
            percentage_list.append(self.get_data_on_percentage(point))

        return percentage_list

    def _on_change_curve(self, text):
        curve_points = self._bezier_dict[text]
        self._scene.undo_stack.push(CurveNodeSwitchUndoCommand(self._scene, self._scene.get_points(), curve_points))
        curve_points = self.curve_as_points()
        self.curveUpdated.emit(curve_points)

    def _on_change_snap(self, flag):
        if not flag:
            self._scene.set_snap(False)
        else:
            self._scene.set_snap(250)

    def _on_curve_updated(self):
        curve_points = self.curve_as_points()
        self.curveUpdated.emit(curve_points)


@theme.mixin
class BezierCurveItem(QGraphicsPathItem, object):
    def __init__(self, parent=None):
        super(BezierCurveItem, self).__init__(parent)

        self._pen = QPen(Qt.green, 2, Qt.SolidLine)
        self._rect = QRectF()
        self._bounding_rect = None
        self._points = list()
        # self._theme = None

        self.setZValue(-1)

    @property
    def points(self):
        return tuple(self._points)

    def boundingRect(self):
        return QRectF(self._bounding_rect or self._rect)

    def paint(self, painter, option, widget):
        self._pen.setColor(self.theme().accent_color if self.theme() else Qt.blue)
        painter.setPen(self._pen)
        painter.drawPath(self.path())

    def update_path(self, points):
        self._points = points
        bezier_path = QPainterPath()
        bezier_path.moveTo(points[0])
        bezier_path.cubicTo(*points[1:])
        self._rect = bezier_path.boundingRect()
        self.setPath(bezier_path)


class CurveNodeItemSignals(QObject):

    curveUpdated = Signal()


@theme.mixin
class CurveNodeItem(QGraphicsItem, object):

    curveUpdated = Signal()
    WIDTH = 10

    def __init__(self, rect=None, parent=None):

        self.signals = CurveNodeItemSignals()
        gradient = QRadialGradient(
            self.WIDTH * 0.75, self.WIDTH * 0.75, self.WIDTH * 0.75, self.WIDTH * 0.75, self.WIDTH * 0.75)
        gradient.setColorAt(0, self.theme().accent_color_6 if self.theme() else QColor.fromRgbF(1, 0.5, 0.01, 1))
        gradient.setColorAt(1, self.theme().accent_color_8 if self.theme() else QColor.fromRgbF(1, 0.6, 0.06, 1))
        self._brush = QBrush(gradient)
        self._brush.setStyle(Qt.RadialGradientPattern)
        self._pen = QPen()
        self._pen.setStyle(Qt.SolidLine)
        self._pen.setWidth(2)
        self._pen.setColor(self.theme().accent_color if self.theme() else QColor(104, 104, 104, 255))
        self._selected_pen = QPen()
        self._selected_pen.setStyle(Qt.SolidLine)
        self._selected_pen.setWidth(3)
        self._selected_pen.setColor(self.theme().accent_color_4 if self.theme() else QColor(67, 255, 163, 255))

        super(CurveNodeItem, self).__init__(parent)

        self._lock_x_pos = False
        self._snap = False
        self._current_pos = None
        self._new_pos = None
        self._line = None
        self._is_point1 = False
        self.set_rect(rect if rect else QRect(0, 0, 10, 10))

        self.setFlags(
            QGraphicsItem.ItemIsMovable | QGraphicsItem.ItemIsSelectable | QGraphicsItem.ItemSendsScenePositionChanges)

    @property
    def snap(self):
        return self._snap

    @snap.setter
    def snap(self, flag):
        self._snap = bool(flag)

    @property
    def lock_x_pos(self):
        return self._lock_x_pos

    @lock_x_pos.setter
    def lock_x_pos(self, value):
        self._lock_x_pos = value

    def boundingRect(self):
        return QRectF(0, 0, 20, 20)

    def paint(self, painter, option, widget):
        painter.setBrush(self._brush)
        painter.setPen(self._selected_pen if self.isSelected() else self._pen)
        painter.drawEllipse(self._rect)

    def mousePressEvent(self, event):
        self._current_pos = self.pos()
        super(CurveNodeItem, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        super(CurveNodeItem, self).mouseMoveEvent(event)
        curve_offset = -self.WIDTH * 0.5
        scale_x = min(max(event.scenePos().x(), curve_offset), self.scene().base_size + curve_offset)
        scale_y = min(max(event.scenePos().y(), curve_offset), self.scene().base_size + curve_offset)
        if self._lock_x_pos:
            scale_x = self._lock_x_pos
        if self._snap is not False:
            scale_x = round((float(scale_x) / self._snap)) * self._snap
            scale_y = round((float(scale_y) / self._snap)) * self._snap
        self._new_pos = QPointF(scale_x, scale_y)
        self.setPos(self._new_pos)
        self.scene().update_curve()
        self.signals.curveUpdated.emit()

    def mouseReleaseEvent(self, event):
        super(CurveNodeItem, self).mouseReleaseEvent(event)

        if not self._new_pos:
            return

        self.scene().undo_stack.push(CurveNodeMoveUndoCommand(self.scene(), self, self._current_pos, self._new_pos))
        self._new_pos = None

    def itemChange(self, change, value):
        if change == self.ItemPositionChange and self.scene():
            new_pos = value
            self._move_line_to_center(new_pos)

        return super(CurveNodeItem, self).itemChange(change, value)

    def set_rect(self, rect):
        self._rect = rect
        self.update()

    def add_line(self, line, is_point1):
        self._line = line
        self._is_point1 = is_point1
        self._move_line_to_center(self.pos())

    def _move_line_to_center(self, new_pos):
        if not self._line:
            return

        x_offset = self._rect.x() - self._rect.width() / 2
        y_offset = self._rect.y() - self._rect.height() / 2
        new_center_pos = QPointF(new_pos.x() - x_offset, new_pos.y() - y_offset)
        p1 = new_center_pos if self._is_point1 else self._line.line().p1()
        p2 = self._line.line().p2() if self._is_point1 else new_center_pos
        self._line.setLine(QLineF(p1, p2))


class CurveNodeScene(QGraphicsScene):

    curveUpdated = Signal()

    def __init__(self, base_rect):
        super(CurveNodeScene, self).__init__()

        self._undo_stack = QUndoStack()
        self._point_objects = list()
        self._control_points = list()
        self._bezier_curve = None
        self._base_size = 300
        self._base_rect = base_rect
        self._theme = None

        self._undo_action = self._undo_stack.createUndoAction(self)
        self._redo_action = self._undo_stack.createRedoAction(self)
        self._undo_action.setShortcut(QKeySequence(Qt.CTRL + Qt.Key_Z))
        self._redo_action.setShortcuts(
            (QKeySequence(Qt.CTRL + Qt.Key_Y),
             QKeySequence(Qt.SHIFT + Qt.Key_Z),
             QKeySequence(Qt.CTRL + Qt.SHIFT + Qt.Key_Z)))

        self.setSceneRect(self._base_rect)
        self._create_grid()
        self._create_grid(10, Qt.black, Qt.DotLine)
        self._create_controls()

    @property
    def bezier_curve(self):
        return self._bezier_curve

    @property
    def undo_stack(self):
        return self._undo_stack

    @property
    def undo_action(self):
        return self._undo_action

    @property
    def redo_action(self):
        return self._redo_action

    @property
    def base_size(self):
        return self._base_size

    @base_size.setter
    def base_size(self, value):
        self._base_size = value

    def keyPressEvent(self, event):
        modifiers = QApplication.keyboardModifiers()
        if type(event) == QKeyEvent:
            if modifiers != Qt.ControlModifier:
                return
            if event.key() == Qt.Key_Z:
                self._undo_action.trigger()
            elif event.key() == Qt.Key_Y:
                self._redo_action.trigger()

    def get_points(self):
        points = list()
        for obj in self._point_objects:
            point = QPointF(obj.pos().x() + 5, obj.pos().y() + 5)
            points.append(point)

        return points

    def set_points(self, points):
        self._point_objects = points

    def set_point_positions(self, positions):
        for i, point in enumerate(positions):
            self._point_objects[i].setPos(point - QPoint(5, 5))
        self.update_curve()

    def set_snap(self, value):
        for point in self._control_points:
            point.snap = value

    def update_curve(self):
        points = self.get_points()
        self._bezier_curve.update_path(points)

    def _create_grid(self, divider=4, color=Qt.darkGray, line=Qt.DashLine):
        pos = float(self._base_size) / divider
        for div in range(divider):
            if div == 0:
                continue
            line_item = QGraphicsLineItem(0, pos * div, self._base_size, pos * div)
            line_item.setZValue(-1)
            line_item.setPen(QPen(color, 1, line))
            self.addItem(line_item)

            line_item = QGraphicsLineItem(pos * div, 0, pos * div, self._base_size)
            line_item.setZValue(-1)
            line_item.setPen(QPen(color, 1, line))
            self.addItem(line_item)

        pen = QPen(Qt.black, 1, Qt.SolidLine)
        rect = QGraphicsRectItem(self._base_rect)
        rect.setZValue(-1)
        self.addItem(rect)

    def _create_controls(self):
        self._control_points = list()
        points = [
            QPoint(0, 0),
            QPoint(self._base_size / 4.0, 0),
            QPoint(self._base_size - (self._base_size / 4.0), self._base_size),
            QPoint(self._base_size, self._base_size)]
        for point in points:
            control_point = CurveNodeItem()
            self.addItem(control_point)
            control_point.setPos(point - QPoint(5, 5))
            self._control_points.append(control_point)
            self._point_objects.append(control_point)
            control_point.signals.curveUpdated.connect(self.curveUpdated.emit)

        self._control_points[0].lock_x_pos = -5.0
        self._control_points[-1].lock_x_pos = (self._base_size - 5.0)
        self._bezier_curve = BezierCurveItem()
        self._bezier_curve.update_path(points)
        self.addItem(self._bezier_curve)

        start_connector = self.addLine(
            QLineF(40, 40, 80, 80), QPen(Qt.white, 1, Qt.DashLine))
        start_connector.setZValue(-2)
        end_connector = self.addLine(
            QLineF(40, 40, 80, 80), QPen(Qt.white, 1, Qt.DashLine))
        end_connector.setZValue(-2)

        self._control_points[0].add_line(start_connector, is_point1=True)
        self._control_points[1].add_line(start_connector, is_point1=False)
        self._control_points[-1].add_line(end_connector, is_point1=False)
        self._control_points[-2].add_line(end_connector, is_point1=True)


class CurveNodeView(QGraphicsView, object):
    def __init__(self, parent=None):
        super(CurveNodeView, self).__init__(parent)

        self.setRenderHints(QPainter.Antialiasing | QPainter.SmoothPixmapTransform | QPainter.HighQualityAntialiasing)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.setViewportUpdateMode(QGraphicsView.FullViewportUpdate)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # self.setStyleSheet('background-color: rgb(60, 60, 60); border: 1px solid rgb(90, 70, 30);')


class CurveNodeMoveUndoCommand(QUndoCommand, object):
    def __init__(self, scene, node, old_position, new_position, description=None, parent=None):
        super(CurveNodeMoveUndoCommand, self).__init__(description or self.__class__.__name__[4:-7], parent)

        self._scene = scene
        self._node = node
        self._old_position = old_position
        self._new_position = new_position

    def redo(self):
        super(CurveNodeMoveUndoCommand, self).redo()
        self._node.setPos(self._new_position)
        self._scene.update_curve()

    def undo(self):
        super(CurveNodeMoveUndoCommand, self).undo()
        self._node.setPos(self._old_position)
        self._scene.update_curve()


class CurveNodeSwitchUndoCommand(QUndoCommand, object):
    def __init__(self, scene, old_positions, new_positions, description=None, parent=None):
        super(CurveNodeSwitchUndoCommand, self).__init__(description or self.__class__.__name__[4:-7], parent)

        self._scene = scene
        self._old_positions = old_positions
        self._new_positions = new_positions

    def redo(self):
        super(CurveNodeSwitchUndoCommand, self).redo()
        self._scene.set_point_positions(self._new_positions)

    def undo(self):
        super(CurveNodeSwitchUndoCommand, self).undo()
        self._scene.set_point_positions(self._old_positions)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library with skinning related functions
"""

from __future__ import print_function, division, absolute_import

import logging
import cStringIO
import traceback

import maya.cmds
import maya.mel

from tpDcc import dcc
from tpDcc.libs.python import python, kdtree, bezier

from tpDcc.dccs.maya.api import skin as api_skin, mathlib as api_mathlib
from tpDcc.dccs.maya.core import decorators, geometry as geo_utils, mesh as mesh_utils, joint as joint_utils
from tpDcc.dccs.maya.core import skin as skin_utils

from tpRigToolkit.tools.rigtoolbox.widgets import library

LOGGER = logging.getLogger('tpRigToolkit-tools-rigtoolbox')


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def apply_smooth_bind_skin(geo=None, show_options=False):
    """
    Initializes joint tool
    """

    out_dict = {'success': False, 'result': None}

    geo_nodes = geo or dcc.selected_nodes_of_type(node_type='transform')
    geo_nodes = python.force_list(geo_nodes)
    valid_geo_nodes = list()
    for geo_node in geo_nodes:
        if not mesh_utils.is_a_mesh(geo_node):
            continue
        valid_geo_nodes.append(geo_node)
    if not valid_geo_nodes:
        out_dict['msg'] = 'No meshes to apply smooth bind skin into found.'
        return out_dict

    try:
        result = skin_utils.apply_smooth_bind(geo=valid_geo_nodes, show_options=show_options)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to apply smooth bind skin: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def apply_rigid_skin(geo=None, show_options=False):

    out_dict = {'success': False, 'result': None}

    geo_nodes = geo or dcc.selected_nodes_of_type(node_type='transform')
    geo_nodes = python.force_list(geo_nodes)
    valid_geo_nodes = list()
    for geo_node in geo_nodes:
        if not mesh_utils.is_a_mesh(geo_node):
            continue
        valid_geo_nodes.append(geo_node)
    if not valid_geo_nodes:
        out_dict['msg'] = 'No meshes to apply rigid bind skin into found.'
        return out_dict

    try:
        result = skin_utils.apply_rigid_skin(geo=valid_geo_nodes, show_options=show_options)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to apply smooth bind skin: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def detach_bind_skin(geo=None, show_options=False):

    out_dict = {'success': False, 'result': list()}

    geo_nodes = geo or dcc.selected_nodes_of_type(node_type='transform')
    geo_nodes = python.force_list(geo_nodes)
    valid_geo_nodes = list()
    for geo_node in geo_nodes:
        if not mesh_utils.is_a_mesh(geo_node):
            continue
        valid_geo_nodes.append(geo_node)
    if not valid_geo_nodes:
        out_dict['msg'] = 'No meshes to apply rigid bind skin into found.'
        return out_dict

    try:
        result = skin_utils.detach_bind_skin(geo=valid_geo_nodes, show_options=show_options)
        out_dict['result'].append(result)
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to apply smooth bind skin: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def open_pain_skin_weights_tool(show_options=False):

    out_dict = {'success': False, 'result': None}

    try:
        result = skin_utils.open_pain_skin_weights_tool(show_options=show_options)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to apply smooth bind skin: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def mirror_skin_weights(mesh=None, show_options=False, **kwargs):

    out_dict = {'success': False, 'result': None}

    transforms = mesh or dcc.selected_nodes_of_type('transform')
    transforms = python.force_list(transforms)
    if not transforms:
        out_dict['msg'] = 'No meshes to mirror skin weights of found.'
        return out_dict

    try:
        result = skin_utils.mirror_skin_weights(mesh=mesh, show_options=show_options, **kwargs)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to mirror skin weights: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def copy_skin_weights(source_mesh=None, target_mesh=None, show_options=False, **kwargs):

    out_dict = {'success': False, 'result': None}

    selection = dcc.selected_nodes_of_type('transform')
    source_transform = source_mesh or (selection[0] if python.index_exists_in_list(selection, 1) else None)
    target_transform = target_mesh or (selection[1] if python.index_exists_in_list(selection, 1) else None)
    if not source_transform or not target_transform:
        out_dict['msg'] = 'Select source mesh and target mesh before executing Copy Skin Weights.'
        return out_dict

    try:
        result = skin_utils.copy_skin_weights(
            source_mesh=source_mesh, target_mesh=target_mesh, show_options=show_options, **kwargs)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to copy skin weights: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def prune_skin_weights(mesh=None, show_options=False, **kwargs):

    out_dict = {'success': False, 'result': None}

    transforms = mesh or dcc.selected_nodes_of_type('transform')
    transforms = python.force_list(transforms)
    if not transforms:
        out_dict['msg'] = 'No meshes to prune skin weights of found.'
        return out_dict

    try:
        result = skin_utils.prune_skin_weights(mesh=mesh, show_options=show_options, **kwargs)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to prune skin weights: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def transfer_uvs_to_skinned_geometry(source_mesh=None, target_mesh=None, use_intermediate_shape=False, **kwargs):

    out_dict = {'success': False, 'result': None}

    selection = dcc.selected_nodes_of_type('transform')
    source_transform = source_mesh or (selection[0] if python.index_exists_in_list(selection, 0) else None)
    target_transform = target_mesh or (selection[1] if python.index_exists_in_list(selection, 1) else None)
    if not source_transform or not target_transform:
        out_dict['msg'] = 'Select source mesh and target mesh before executing Transfers UVs to skinned geometry.'
        return out_dict

    try:
        result = skin_utils.transfer_uvs_to_skinned_geometry(
            source_mesh=source_mesh, target_mesh=target_mesh, use_intermediate_shape=use_intermediate_shape, **kwargs)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to transfers UVs to skinned geometry: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def freeze_skinned_mesh(skinned_mesh, **kwargs):

    out_dict = {'success': False, 'result': list()}

    meshes = skinned_mesh or dcc.selected_nodes_of_type('transform')
    meshes = python.force_list(meshes)
    if not meshes:
        return False

    if kwargs.pop('auto_assign_labels', False):
        joint_utils.auto_assign_labels_to_mesh_influences(
            meshes, input_left=kwargs.pop('left_side_label', None),
            input_right=kwargs.pop('right_side_label', None), check_labels=True)

    percentage = 100.0 / len(meshes)

    for i, mesh in enumerate(meshes):
        library.Command.progressCommand.emit(percentage * (i + 1), 'Cleaning Skinned Mesh: {}'.format(mesh))
        try:
            skin_cluster_name = skin_utils.find_related_skin_cluster(mesh)
            if not skin_cluster_name:
                continue
            attached_joints = maya.cmds.skinCluster(skin_cluster_name, query=True, inf=True)
            mesh_shape_name = maya.cmds.listRelatives(mesh, shapes=True)[0]
            out_influences_array = api_skin.get_skin_weights(skin_cluster_name, mesh_shape_name)
            maya.cmds.skinCluster(mesh_shape_name, edit=True, unbind=True)
            maya.cmds.delete(mesh, ch=True)
            maya.cmds.makeIdentity(mesh, apply=True)
            new_skin_cluster_name = maya.cmds.skinCluster(
                attached_joints, mesh, toSelectedBones=True, bindMethod=0, normalizeWeights=True)[0]
            api_skin.set_skin_weights(new_skin_cluster_name, mesh_shape_name, out_influences_array)
            out_dict['result'].append(mesh)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to freeze skinned meshes: "{}" | {}'.format(meshes, exc)
            return out_dict

    dcc.select_node(meshes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def combine_skinned_meshes(meshes=None):

    out_dict = {'success': False, 'result': None}

    meshes = meshes or dcc.selected_nodes_of_type('transform')
    if not meshes:
        out_dict['msg'] = 'No meshes to combine found.'
        return out_dict

    try:
        result = skin_utils.combine_skinned_meshes(meshes)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to transfers UVs to skinned geometry: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def extract_skinned_selected_components(selected_components=None, **kwargs):

    out_dict = {'success': False, 'result': None}

    components = selected_components or dcc.selected_nodes(flatten=True)
    components = python.force_list(components)
    if not components:
        out_dict['msg'] = 'No components to extract from found.'
        return out_dict

    try:
        result = skin_utils.extract_skinned_selected_components(components, **kwargs)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to transfers UVs to skinned geometry: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def delete_unused_influences(skinned_objects=None):

    out_dict = {'success': False, 'result': list()}

    skinned_objects = skinned_objects or dcc.selected_nodes()
    skinned_objects = python.force_list(skinned_objects)
    if not skinned_objects:
        out_dict['msg'] = 'No components to extract from found.'
        return out_dict

    percentage = 100.0 / len(skinned_objects)

    for i, mesh in enumerate(skinned_objects):
        try:
            library.Command.progressCommand.emit(percentage * (i + 1), 'Deleting unused influences: {}'.format(mesh))
            skin_cluster_name = skin_utils.find_related_skin_cluster(mesh)
            if not skin_cluster_name:
                shape = maya.cmds.listRelatives(mesh, shapes=True) or None
                if shape:
                    LOGGER.warning(
                        'Impossible to delete unused influences because mesh "{}" '
                        'has no skin cluster attached to it!'.format(mesh))
                continue

            attached_joints = maya.cmds.skinCluster(skin_cluster_name, query=True, influence=True)
            weighted_joints = maya.cmds.skinCluster(skin_cluster_name, query=True, weightedInfluence=True)

            non_influenced = list()
            for attached in attached_joints:
                if attached in weighted_joints:
                    continue
                non_influenced.append(attached)

            for joint in non_influenced:
                maya.cmds.skinCluster(skin_cluster_name, edit=True, removeInfluence=joint)
            out_dict['result'].append(mesh)
        except Exception as exc:
            out_dict['msg'] = 'Was not possible to delete unused influences: "{}" | {}'.format(skinned_objects, exc)
            return out_dict

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def restore_to_bind_pose(skinned_mesh=None):

    out_dict = {'success': False, 'result': None}

    skinned_meshes = skinned_mesh or dcc.selected_nodes_of_type('transform')
    skinned_meshes = python.force_list(skinned_meshes)
    if not skinned_meshes:
        skinned_meshes = list()
        meshes = dcc.list_nodes(node_type='mesh')
        for mesh in meshes:
            skinned_meshes.append(maya.cmds.listRelatives(mesh, parent=True)[0])
    if not skinned_meshes:
        out_dict['msg'] = 'No skinned meshes to restore bind poses of found.'
        return out_dict

    try:
        result = skin_utils.restore_to_bind_pose(skinned_mesh)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to restore skinned mesh to bind pose: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def remove_bind_poses():

    out_dict = {'success': False, 'result': None}

    try:
        result = skin_utils.remove_all_bind_poses_in_scene()
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to remove all bind poses from current scene: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


def hammer_vertices(vertices_to_hammer=None, return_as_list=True):

    out_dict = {'success': False, 'result': None}

    try:
        result = skin_utils.hammer_vertices(vertices_to_hammer, return_as_list=return_as_list)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to hammer vertices: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
def average_vertices_weights(selection, use_distance, curve_weight_points=None):
    """
    Generates an average weight from all selected vertices to apply to the last selected vertex
    :param selection: list<Vertex>, list of vertices to average
    :param use_distance:
    :param curve_weight_points:
    :return:
    """

    out_dict = {'success': False, 'result': None}

    total_vertices = len(selection)
    if total_vertices < 2:
        out_dict['msg'] = 'Not enough vertices selected! Select a minimum of 2 vertices.'
        return out_dict

    obj = selection[0]
    if '.' in selection[0]:
        obj = selection[0].split('.')[0]

    is_edge_selection = False
    if '.e[' in selection[0]:
        is_edge_selection = True

    skin_cluster_name = skin_utils.find_related_skin_cluster(obj)
    maya.cmds.setAttr('{0}.envelope'.format(skin_cluster_name), 0)

    try:
        maya.cmds.skinCluster(obj, edit=True, normalizeWeights=True)
        if total_vertices == 2 or is_edge_selection:
            base_list = [selection]
            if is_edge_selection:
                base_list = mesh_utils.edges_to_smooth(edges_list=selection)

            percentage = 99.0 / len(base_list)

            for i, vert_list in enumerate(base_list):
                library.Command.progressCommand.emit(
                    percentage * (i + 1), 'Pass 1: Averaging vertex weights: {}'.format(i))

                start = vert_list[0]
                end = vert_list[-1]
                surface = start.split('.')[0]
                obj_type = maya.cmds.objectType(surface)
                if obj_type == 'transform':
                    shape = maya.cmds.listRelatives(surface, shapes=True)
                    if shape:
                        obj_type = maya.cmds.objectType(shape[0])

                poly = False
                if obj_type == 'mesh':
                    poly = True
                    order = mesh_utils.find_shortest_vertices_path_between_vertices(vert_list)
                    added = 0.0
                    amount = len(order) + 1
                    total_distance = api_mathlib.distance_between_nodes(order[-1], end)
                elif obj_type == 'nurbsSurface':
                    order, total_distance = geo_utils.find_shortest_path_between_surface_cvs(
                        vert_list, return_total_distance=True)
                    added = -2.0
                    amount = len(order) + 1
                elif obj_type == 'lattice':
                    order, total_distance = geo_utils.find_shortest_path_between_lattice_cvs(
                        vert_list, return_total_distance=True)
                    added = -2.0
                    amount = len(order) + 1
                else:
                    numbers = [int(start.split("[")[-1].split("]")[0]), int(end.split("[")[-1].split("]")[0])]
                    range_list = range(min(numbers), max(numbers) + 1)
                    amount = len(range_list)
                    added = -1.0
                    order = list()
                    total_distance = 0.0
                    for j, num in enumerate(range_list):
                        cv = '{}.cv[{}]'.format(surface, num)
                        order.append(cv)
                        if j == 0:
                            continue
                        total_distance += api_mathlib.distance_between_nodes(order[j - 1], cv)
                if not order:
                    return

                list_bone_influences = maya.cmds.skinCluster(obj, query=True, inf=True)
                weights_start = maya.cmds.skinPercent(skin_cluster_name, start, query=True, v=True)
                weights_end = maya.cmds.skinPercent(skin_cluster_name, end, query=True, v=True)

                lengths = list()
                if use_distance:
                    for j, vertex in enumerate(order):
                        if j == 0:
                            length = api_mathlib.distance_between_nodes(start, vertex)
                        else:
                            length = api_mathlib.distance_between_nodes(order[j - 1], vertex)
                        if poly:
                            total_distance += length
                        lengths.append(length)

                percentage = float(1.0) / (amount + added)
                current_length = 0.0

                for index, vertex in enumerate(order):
                    library.Command.progressCommand.emit(
                        percentage * (index + 1), 'Pass 2: Averaging vertex weights: {}'.format(index))
                    if use_distance:
                        current_length += lengths[index]
                        current_percentage = (current_length / total_distance)
                    else:
                        current_percentage = index * percentage
                        if poly:
                            current_percentage = (index + 1) * percentage
                    if curve_weight_points:
                        current_percentage = bezier.get_data_on_percentage(current_percentage, curve_weight_points)

                    new_weight_list = list()
                    for j, weight in enumerate(weights_start):
                        value1 = weights_end[j] * current_percentage
                        value2 = weights_start[j] * (1 - current_percentage)
                        new_weight_list.append((list_bone_influences[j], value1 + value2))

                    maya.cmds.skinPercent(skin_cluster_name, vertex, transformValue=new_weight_list)

                maya.cmds.select([start, end], r=True)

                maya.cmds.setAttr('{}.envelope'.format(skin_cluster_name), 1)
                maya.cmds.select(vert_list, replace=True)
                maya.cmds.refresh()
                library.Command.progressCommand.emit(
                    percentage * i, 'Pass 3: Updating skin cluster for vertex: {}'.format(i))
                maya.cmds.setAttr('{}.envelope'.format(skin_cluster_name), 0)
        else:
            last_selected = selection[-1]
            point_list = [x for x in selection if x != last_selected]
            mesh_name = last_selected.split('.')[0]

            list_joint_influences = maya.cmds.skinCluster(mesh_name, query=True, weightedInfluence=True)
            influence_size = len(list_joint_influences)

            temp_vertex_joints = list()
            temp_vertex_weights = list()
            for pnt in point_list:
                for jnt in range(influence_size):
                    point_weights = maya.cmds.skinPercent(
                        skin_cluster_name, pnt, transform=list_joint_influences[jnt], query=True, value=True)
                    if point_weights < 0.000001:
                        continue
                    temp_vertex_joints.append(list_joint_influences[jnt])
                    temp_vertex_weights.append(point_weights)

            total_values = 0.0
            average_values = list()
            clean_list = list()
            for i in temp_vertex_joints:
                if i not in clean_list:
                    clean_list.append(i)

            for i in range(len(clean_list)):
                working_value = 0.0
                for j in range(len(temp_vertex_joints)):
                    if not temp_vertex_joints[j] == clean_list[i]:
                        continue
                    working_value += temp_vertex_weights[j]
                num_points = len(point_list)
                average_values.append(working_value / num_points)
                total_values += average_values[i]

            summary = 0
            for value in range(len(average_values)):
                temp_value = average_values[value] / total_values
                average_values[value] = temp_value
                summary += average_values[value]

            cmd = cStringIO.StringIO()
            cmd.write('maya.cmds.skinPercent("%s","%s", transformValue=[' % (skin_cluster_name, last_selected))

            for count, skin_joint in enumerate(clean_list):
                cmd.write('("%s", %s)' % (skin_joint, average_values[count]))
                if not count == len(clean_list) - 1:
                    cmd.write(', ')
            cmd.write('])')
            eval(cmd.getvalue())
    except Exception:
        out_dict['msg'] = 'Was not possible to average vertex weights: {}'.format(traceback.format_exc())
        maya.cmds.setAttr('{}.envelope'.format(skin_cluster_name), 1)
        return out_dict
    finally:
        maya.cmds.setAttr('{}.envelope'.format(skin_cluster_name), 1)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def move_skin_weights(source_joint=None, target_joint=None, mesh=None):

    out_dict = {'success': False, 'result': None}

    selection = dcc.selected_nodes()
    source_joint = source_joint or (selection[0] if python.index_exists_in_list(selection, 0) else None)
    if not source_joint:
        out_dict['msg'] = 'No source joint found to move skin weights from.'
        return out_dict

    target_joint = target_joint or (selection[1] if python.index_exists_in_list(selection, 1) else None)
    if not target_joint:
        out_dict['msg'] = 'No target joint found to move skin weights to.'
        return out_dict

    mesh = mesh or (selection[2] if python.index_exists_in_list(selection, 2) else None)
    if not mesh:
        out_dict['msg'] = 'No mesh with skinning information found'
        return out_dict

    try:
        result = skin_utils.move_skin_weights(source_joint, target_joint, mesh)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to move skin weights: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def swap_skin_weights(source_joint=None, target_joint=None, mesh=None):

    out_dict = {'success': False, 'result': None}

    selection = dcc.selected_nodes()
    source_joint = source_joint or (selection[0] if python.index_exists_in_list(selection, 0) else None)
    if not source_joint:
        out_dict['msg'] = 'No source joint found to swap skin weights from.'
        return out_dict

    target_joint = target_joint or (selection[1] if python.index_exists_in_list(selection, 1) else None)
    if not target_joint:
        out_dict['msg'] = 'No target joint found to swap skin weights to.'
        return out_dict

    mesh = mesh or (selection[2] if python.index_exists_in_list(selection, 2) else None)
    if not mesh:
        out_dict['msg'] = 'No mesh with skinning information found'
        return out_dict

    try:
        result = skin_utils.swap_skin_weights(source_joint, target_joint, mesh)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to swap skin weights: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def select_influencing_joints(mesh_node=None):

    out_dict = {'success': False, 'result': None}

    mesh_nodes = mesh_node or dcc.selected_nodes_of_type(node_type='transform')
    mesh_nodes = python.force_list(mesh_nodes)
    mesh_node = mesh_nodes[0] if mesh_nodes else None
    if not mesh_node:
        out_dict['msg'] = 'No mesh selected to retrieve influences of'
        return out_dict

    try:
        result = skin_utils.select_influencing_joints(mesh_node)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to select influencing joints: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def select_influence_components(joint_nodes=None, mesh_node=None):

    out_dict = {'success': False, 'result': None}

    try:
        result = skin_utils.select_influence_vertices(joint_nodes=joint_nodes, mesh_node=mesh_node)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to select influencing joints: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def delete_influences(skinned_objects=None, influences_to_remove=None, fast=True):
    """
    Deletes given influences from given meshes and stores the unbind influences weights into other influences
    :param skinned_objects: list(str), meshes which joints need to be removed
    :param influences_to_remove: list(str), list of joints that need to be removed
    :param fast: list(str), bool, Whether to do the deletion using quick unbind method or not
    :return: bool
    """

    if fast:
        valid_deletion = unbind_influences_quick(
            skinned_objects=skinned_objects, influences_to_unbind=influences_to_remove, delete=True)
    else:
        valid_deletion = unbind_influences(
            skinned_objects=skinned_objects, influences_to_unbind=influences_to_remove, delete=True
        )

    return valid_deletion


@decorators.undo
def unbind_influences_quick(skinned_objects=None, influences_to_unbind=None, delete=False):
    """
    Unbind given influences from given meshes and stores the unbind influences weights into other influences
    The weights reassignation is handled by Maya
    :param skinned_objects: list(str), meshes which joints need to be removed
    :param influences_to_unbind: list(str), list of joints that need to be unbind
    :param delete: bool, Whether or not to delete unbind influences after unbind process is completed
    :return: bool
    """

    selected_transforms = dcc.selected_nodes_of_type('transform')
    selected_joints = dcc.selected_nodes_of_type('joint')
    influences_to_unbind = influences_to_unbind or selected_joints
    if not skinned_objects:
        skinned_objects = [xform for xform in selected_transforms if xform not in selected_joints]
    if not skinned_objects or not influences_to_unbind:
        return False
    skinned_objects = python.force_list(skinned_objects)
    influences_to_unbind = python.force_list(influences_to_unbind)
    influences_to_unbind_short = [dcc.node_short_name(joint_node) for joint_node in influences_to_unbind]

    skin_clusters = list()
    skin_percentage = 100.0 / len(skinned_objects)

    for i, skin_object in enumerate(skinned_objects):

        library.Command.progressCommand.emit(skin_percentage * (i + 1), 'Unbinding Influence: {}'.format(i))

        skin_cluster_name = skin_utils.find_related_skin_cluster(skin_object)
        if not skin_cluster_name:
            continue
        joints_attached = maya.cmds.skinCluster(skin_cluster_name, query=True, inf=True)

        influence_verts = skin_utils.get_influence_vertices(influences_to_unbind, skin_object)
        if not influence_verts:
            continue

        joints = list()
        for joint_to_remove in influences_to_unbind_short:
            if joint_to_remove not in joints_attached:
                continue
            joints.append((joint_to_remove, 0.0))

        maya.cmds.select(influence_verts, replace=True)
        maya.cmds.skinPercent(skin_cluster_name, transformValue=joints, normalize=True)

        skin_clusters.append(skin_cluster_name)

    for skin_cluster in skin_clusters:
        joints_attached = maya.cmds.skinCluster(skin_cluster, query=True, inf=True)
        for jnt in influences_to_unbind_short:
            if jnt not in joints_attached:
                continue
            maya.cmds.skinCluster(skin_cluster, edit=True, removeInfluence=jnt)

    if delete:
        for joint_to_remove in influences_to_unbind:
            child_joints = maya.cmds.listRelatives(joint_to_remove, children=True)
            parent = maya.cmds.listRelatives(joint_to_remove, parent=True)
            if not child_joints:
                continue
            if not parent:
                maya.cmds.parent(child_joints, world=True)
                continue
            maya.cmds.parent(child_joints, parent)
        maya.cmds.delete(influences_to_unbind)

    return True


@decorators.undo
def unbind_influences(skinned_objects=None, influences_to_unbind=None, delete=False, use_parent=True):
    """
    Unbind given influences from given meshes and stores the unbind influences weights into other influences

    :param skinned_objects: list(str), meshes which joints need to be removed
    :param influences_to_unbind: list(str), list of joints that need to be unbind
    :param delete: bool, Whether or not to delete unbind influences after unbind process is completed
    :param use_parent: bool, If True, removed influences weights will be stored on its parent; if False it will look
        for the closest joint using a point cloud system
    :return: bool
    """

    selected_transforms = dcc.selected_nodes_of_type('transform')
    selected_joints = dcc.selected_nodes_of_type('joint')
    influences_to_unbind = influences_to_unbind or selected_joints
    if not skinned_objects:
        skinned_objects = [xform for xform in selected_transforms if xform not in selected_joints]
    if not skinned_objects or not influences_to_unbind:
        return False
    skinned_objects = python.force_list(skinned_objects)
    influences_to_unbind = python.force_list(influences_to_unbind)
    influences_to_unbind_short = [dcc.node_short_name(joint_node) for joint_node in influences_to_unbind]

    skin_clusters = list()
    skin_percentage = 100.0 / len(skinned_objects)

    for skin_index, skin_object in enumerate(skinned_objects):
        skin_cluster_name = skin_utils.find_related_skin_cluster(skin_object)
        if not skin_cluster_name:
            continue
        joints_attached = maya.cmds.skinCluster(skin_cluster_name, query=True, inf=True)

        if not use_parent:
            for joint_to_remove in influences_to_unbind_short:
                if joint_to_remove in joints_attached:
                    joints_attached.remove(joint_to_remove)

        source_positions = list()
        source_joints = list()
        for joint_attached in joints_attached:
            pos = maya.cmds.xform(joint_attached, query=True, worldSpace=True, t=True)
            source_positions.append(pos)
            source_joints.append([joint_attached, pos])

        source_kdtree = kdtree.KDTree.construct_from_data(source_positions)

        joint_percentage = skin_percentage / len(influences_to_unbind)
        for joint_index, jnt in enumerate(influences_to_unbind):
            jnt1 = jnt
            if use_parent:
                jnt2 = maya.cmds.listRelatives(jnt, parent=True)
                jnt2 = jnt2[0] if jnt2 else None
                if jnt2 is None:
                    remove_pos = maya.cmds.xform(jnt, query=True, worldSpace=True, t=True)
                    points = source_kdtree.query(query_point=remove_pos, t=1)
                    for index, position in enumerate(source_joints):
                        if position[1] != points[0]:
                            continue
                        jnt2 = position[0]
            else:
                remove_pos = maya.cmds.xform(jnt, query=True, worldSpace=True, t=True)
                points = source_kdtree.query(query_point=remove_pos, t=True)
                for index, position in enumerate(source_joints):
                    if position[1] != points[0]:
                        continue
                    jnt2 = position[0]

            move_skin_weights(jnt1, jnt2, skin_object)

            library.Command.progressCommand.emit(
                ((joint_index + 1) * joint_percentage) + (skin_index * skin_percentage), 'Unbinding Influence: {}')

        skin_clusters.append(skin_cluster_name)

    for skin_cluster in skin_clusters:
        joints_attached = maya.cmds.skinCluster(skin_cluster, query=True, inf=True)
        for jnt in influences_to_unbind_short:
            if jnt not in joints_attached:
                continue
            maya.cmds.skinCluster(skin_cluster, edit=True, removeInfluence=jnt)

    if delete:
        for joint_to_remove in influences_to_unbind:
            child_joints = maya.cmds.listRelatives(joint_to_remove, children=True)
            parent = maya.cmds.listRelatives(joint_to_remove, parent=True)
            if not child_joints:
                continue
            if not parent:
                maya.cmds.parent(child_joints, world=True)
                continue
            maya.cmds.parent(child_joints, parent)
        maya.cmds.delete(influences_to_unbind)

    return True


@decorators.repeat_static_command(__name__, skip_arguments=True)
def br_smooth_weights():
    """
    Executes Brave Rabbit Smooth Weights command if available
    :return: bool
    """

    if not dcc.is_plugin_loaded('brSmoothWeights.mll'):
        return False

    maya.mel.eval('brSmoothWeightsToolCtx')

    return True


@decorators.repeat_static_command(__name__, skip_arguments=True)
def br_smooth_weights_options():
    """
    Opens Brave Rabbit Smooth Weights tool if available
    :return: bool
    """

    if not dcc.is_plugin_loaded('brSmoothWeights.mll'):
        return False

    maya.mel.eval('brSmoothWeightsToolCtx; toolPropertyWindow;')

    return True


@decorators.repeat_static_command(__name__, skip_arguments=True)
def br_transfer_weights():
    """
    Executes Brave Rabbit Transfer Weights command if available
    :return: bool
    """

    plugin_name = 'brSmoothWeights.mll'
    if not dcc.is_plugin_loaded(plugin_name):
        dcc.load_plugin(plugin_name, quiet=True)
        if not dcc.is_plugin_loaded(plugin_name):
            return False

    maya.mel.eval('brTransferWeightsToolCtx')

    return True


@decorators.repeat_static_command(__name__, skip_arguments=True)
def br_transfer_weights_options():
    """
    Opens Brave Rabbit Transfer Weights tool if available
    :return: bool
    """

    plugin_name = 'brSmoothWeights.mll'
    if not dcc.is_plugin_loaded(plugin_name):
        dcc.load_plugin(plugin_name, quiet=True)
        if not dcc.is_plugin_loaded(plugin_name):
            return False
        return False

    maya.mel.eval('brTransferWeightsToolCtx; toolPropertyWindow;')

    return True


@decorators.repeat_static_command(__name__, skip_arguments=True)
def ng_skin_tools():
    """
    Opens ngSkinTools tool if available
    First, we try to open ngSkinTools2 and if it is not available, we launch ngSkinTools1
    :return: bool
    """

    plugin_names = ['ngSkinTools2.mll', 'ngSkinTools.mll']
    for plugin_name in plugin_names:
        if not dcc.is_plugin_loaded(plugin_name):
            dcc.load_plugin(plugin_name, quiet=True)
        if dcc.is_plugin_loaded(plugin_name):
            if plugin_name == 'ngSkinTools2.mll':
                try:
                    import ngSkinTools2
                    ngSkinTools2.open_ui()
                    return True
                except Exception as exc:
                    LOGGER.warning(
                        'Impossible to launch ngSkinTool2. Trying to launch ngSkinTools1: "{}"'.format(exc))
                    continue
            else:
                try:
                    from ngSkinTools.ui.mainwindow import MainWindow
                    MainWindow.open()
                    return True
                except Exception as exc:
                    'Impossible to launch ngSkinTool1: "{}"'.format(exc)
                    continue

    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains constants definitions used by tpRigToolkit-tools-rigtoolbox
"""

from __future__ import print_function, division, absolute_import

TOOL_ID = 'tpRigToolkit-tools-rigtoolbox'



from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Signal, QSize, QEvent
from Qt.QtWidgets import QWidget, QDockWidget, QLabel, QLineEdit, QToolButton, QGroupBox

from tpDcc.managers import resources
from tpDcc.libs.qt.widgets import layouts


class DockWidget(QDockWidget, object):
    closed = Signal(object)

    def __init__(self, name, parent=None):
        super(DockWidget, self).__init__(parent=parent)

        self.setWindowTitle(name)
        self.setFeatures(QDockWidget.AllDockWidgetFeatures)
        self.setAllowedAreas(
            Qt.BottomDockWidgetArea | Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea | Qt.TopDockWidgetArea)
        self.setObjectName(name)
        self.setTitleBarWidget(DockTitleBar(self))
        self.setFloating(False)

    def closeEvent(self, event):
        self.closed.emit(self)
        event.accept()

    def add_button(self, button):
        self.titleBarWidget().add_button(button)


class DockTitleBar(QWidget, object):
    def __init__(self, dock_widget, renamable=False):
        super(DockTitleBar, self).__init__(dock_widget)

        self._renamable = renamable

        main_layout = layouts.HorizontalLayout(spacing=1, margins=(0, 0, 0, 1))
        self.setLayout(main_layout)

        self._buttons_box = QGroupBox('')
        self._buttons_box.setObjectName('Docked')
        self._buttons_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._buttons_layout.setSpacing(1)
        self._buttons_layout.setMargin(2)
        self._buttons_box.setLayout(self._buttons_layout)
        main_layout.addWidget(self._buttons_box)

        self._title_label = QLabel(self)
        self._title_label.setStyleSheet('background: transparent')
        self._title_edit = QLineEdit(self)
        self._title_edit.setVisible(False)

        self._button_size = QSize(15, 15)

        self._dock_btn = QToolButton(self)
        self._dock_btn.setIcon(resources.icon('restore_window', theme='color'))
        self._dock_btn.setMaximumSize(self._button_size)
        self._dock_btn.setAutoRaise(True)
        self._close_btn = QToolButton(self)
        self._close_btn.setIcon(resources.icon('close_window', theme='color'))
        self._close_btn.setMaximumSize(self._button_size)
        self._close_btn.setAutoRaise(True)

        self._buttons_layout.addSpacing(2)
        self._buttons_layout.addWidget(self._title_label)
        self._buttons_layout.addWidget(self._title_edit)
        self._buttons_layout.addStretch()
        self._buttons_layout.addSpacing(5)
        self._buttons_layout.addWidget(self._dock_btn)
        self._buttons_layout.addWidget(self._close_btn)

        self._buttons_box.mouseDoubleClickEvent = self.mouseDoubleClickEvent
        self._buttons_box.mousePressEvent = self.mousePressEvent
        self._buttons_box.mouseMoveEvent = self.mouseMoveEvent
        self._buttons_box.mouseReleaseEvent = self.mouseReleaseEvent

        dock_widget.featuresChanged.connect(self._on_dock_features_changed)
        self._title_edit.editingFinished.connect(self._on_finish_edit)
        self._dock_btn.clicked.connect(self._on_dock_btn_clicked)
        self._close_btn.clicked.connect(self._on_close_btn_clicked)

        self._on_dock_features_changed(dock_widget.features())
        self.set_title(dock_widget.windowTitle())
        dock_widget.installEventFilter(self)
        dock_widget.topLevelChanged.connect(self._on_change_floating_style)

    @property
    def renamable(self):
        return self._renamable

    @renamable.setter
    def renamable(self, flag):
        self._renamable = flag

    def eventFilter(self, obj, event):
        try:
            if event.type() == QEvent.WindowTitleChange:
                self.set_title(obj.windowTitle())
            return super(DockTitleBar, self).eventFilter(obj, event)
        except Exception as exc:
            event.accept()
            return True

    def mouseMoveEvent(self, event):
        event.ignore()

    def mousePressEvent(self, event):
        event.ignore()

    def mouseReleaseEvent(self, event):
        event.ignore()

    def mouseDoubleClickEvent(self, event):
        if event.pos().x() <= self._title_label.width() and self._renamable:
            self._start_edit()
        else:
            super(DockTitleBar, self).mouseDoubleClickEvent(event)

    def update(self, *args, **kwargs):
        self._on_change_floating_style(self.parent().isFloating())
        super(DockTitleBar, self).update(*args, **kwargs)

    def set_title(self, title):
        self._title_label.setText(title)
        self._title_edit.setText(title)

    def add_button(self, button):
        button.setAutoRaise(True)
        button.setMaximumSize(self._button_size)
        self._buttons_layout.insertWidget(5, button)

    def _start_edit(self):
        self._title_label.hide()
        self._title_edit.show()
        self._title_edit.setFocus()

    def _finish_edit(self):
        self._title_edit.hide()
        self._title_label.show()
        self.parent().setWindowTitle(self._title_edit.text())

    def _on_dock_features_changed(self, features):
        if not features & QDockWidget.DockWidgetVerticalTitleBar:
            self._close_btn.setVisible(features & QDockWidget.DockWidgetClosable)
            self._dock_btn.setVisible(features & QDockWidget.DockWidgetFloatable)
        else:
            raise ValueError('Vertical title bar is not supported!')

    def _on_finish_edit(self):
        self._finish_edit()

    def _on_dock_btn_clicked(self):
        self.parent().setFloating(not self.parent().isFloating())

    def _on_close_btn_clicked(self):
        self.parent().toggleViewAction().setChecked(False)
        self.parent().close()

    def _on_change_floating_style(self, state):
        pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base class for tpRigToolbox widgets
"""

from __future__ import print_function, division, absolute_import

import time
import inspect
from functools import partial

from Qt.QtCore import Qt, Signal, QPoint, QSize, QEvent
from Qt.QtWidgets import QSizePolicy, QWidget, QFrame, QMenu, QAction, QWhatsThis
from Qt.QtGui import QCursor, QColor, QPainter, QBrush, QPolygon

from tpDcc.managers import resources
from tpDcc.libs.python import python
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, label, buttons, accordion, balloon

from tpRigToolkit.tools.rigtoolbox.widgets import library, info


class BaseRigToolBoxWidget(base.BaseFrame, object):

    emitInfo = Signal(str)
    emitWarning = Signal(str)
    emitError = Signal(str)
    setInfo = Signal(str, str, str)

    def __init__(self, title, parent=None):

        self._title = title

        super(BaseRigToolBoxWidget, self).__init__(parent=parent)

    @property
    def title(self):
        return self._title

    def get_main_layout(self):
        main_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        main_layout.setAlignment(Qt.AlignTop)

        return main_layout

    def ui(self):
        super(BaseRigToolBoxWidget, self).ui()
        self._content_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

        self.main_layout.addLayout(self._content_layout)


class CommandRigToolBoxWidget(BaseRigToolBoxWidget, object):
    def __init__(self, title, commands_data, controller, parent=None):

        self._commands_data = commands_data or dict()
        self._widgets = dict()
        self._buttons = list()
        self._current_action = None
        self._controller = controller
        self._controller_functions_mapping = dict()

        super(CommandRigToolBoxWidget, self).__init__(title=title, parent=parent)

        if self._controller:
            predicate = inspect.ismethod if python.is_python2() else inspect.isfunction
            controller_functions = inspect.getmembers(self._controller.__class__, predicate=predicate)
            for controller_fn_data in controller_functions:
                if controller_fn_data[0] in ['__init__']:
                    continue
                controller_fn = getattr(self._controller, controller_fn_data[0])
                if not controller_fn:
                    continue
                self._controller_functions_mapping[controller_fn_data[0]] = controller_fn

        added_categories = list()
        for name, data in self._commands_data.items():
            categories = data.get('categories', list())
            for category in categories:
                if not category or category in added_categories:
                    continue
                category_widget = self._add_category(category)
                if not category_widget:
                    continue
                self._accordion.add_item(category, category_widget)
                added_categories.append(category)

    def ui(self):
        super(CommandRigToolBoxWidget, self).ui()

        self._info_widget = info.InfoMessage(parent=self)
        self._info_frame = base.BaseFrame(parent=self)
        self._info_label = label.BaseLabel(parent=self).strong()
        self._info_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)
        info_frame_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        info_frame_layout.addStretch()
        info_frame_layout.addWidget(self._info_label)
        info_frame_layout.addStretch()
        self._info_frame.main_layout.addLayout(info_frame_layout)
        self._info_frame.setVisible(False)

        self._accordion = accordion.AccordionWidget(parent=self)
        self._content_layout.addWidget(self._accordion)

        self.main_layout.addWidget(self._info_widget)
        self.main_layout.addWidget(self._info_frame)

    def eventFilter(self, obj, event):

        if isinstance(obj, QMenu):
            if event.type() == QEvent.MouseMove:
                action = obj.actionAt(event.pos())
                if action:
                    if action != self._current_action:
                        self._current_action = action
                        name = action.property('name') or action.text()
                        description = action.property('description') or ''
                        instructions = action.property('instructions') or ''
                        self._set_info(name, description, instructions)

                        self.setInfo.emit(name, description, instructions)
            if event.type() == QEvent.HoverLeave or event.type() == QEvent.Leave:
                self._set_info('', '', '')
                self._current_action = None
        else:
            if event.type() == QEvent.HoverEnter:
                name = obj.property('name') or obj.text()
                description = obj.property('description') or ''
                instructions = obj.property('instructions') or ''
                self._set_info(name, description, instructions)
            elif event.type() == QEvent.HoverLeave:
                self._set_info('', '', '')

        return super(BaseRigToolBoxWidget, self).eventFilter(obj, event)

    def _add_category(self, category_name):
        if not self._controller:
            return False

        commands_to_add = list()

        for command_name, command_data in self._commands_data.items():
            command_function = self._controller_functions_mapping.get(command_name, None)
            if not command_data or command_name not in self._controller_functions_mapping:
                continue
            command_categories = command_data.get('categories', list())
            if category_name not in command_categories:
                continue
            options = list()
            command_actions = command_data.get('options', dict())
            for command_option_name, command_option_data in command_actions.items():
                command_option_function = self._controller_functions_mapping.get(command_option_name, None)
                if not command_option_function:
                    continue
                option_data = command_option_data.copy()
                option_data['fn'] = command_option_function
                options.append(option_data)

            new_command_data = command_data.copy()
            new_command_data.pop('categories')       # categories is not part of the command creation
            new_command_data.pop('options', None)
            new_command = self._create_button(fn=command_function, settings=options, **new_command_data)

            is_available = self._check_command_availability(command_name)
            if not is_available:
                new_command.setToolTip('Command "{}" is not available!'.format(command_name))
                new_command.setEnabled(False)

            commands_to_add.append(new_command)

        if not commands_to_add:
            return False

        category_widget = QWidget()
        category_layout = layouts.FlowLayout()
        category_layout.setAlignment(Qt.AlignLeft)
        category_widget.setLayout(category_layout)

        for command in commands_to_add:
            category_layout.addWidget(command)

        return category_widget

    def _check_command_availability(self, command_name):
        """
        Internal function that returns whether or not given command is available
        Should be override in custom widgets to enable/disable commands
        By default, all commands are available
        :param command_name: str
        :return: bool
        """

        return True

    def _create_button(
            self, name=None, icon=None, fn=None, status=None, tooltip=None, description=None, instructions=None,
            whats_this='', settings=None, settings_fn=None, info_message='', error_message='', widgets=None):
        """
        Internal function to easily create buttons for the box widget
        :param name:
        :param icon:
        :param fn:
        :param status:
        :param tooltip:
        :return:
        """

        name = name or ''
        new_btn = CommandButton(parent=self).small()
        if icon:
            if python.is_string(icon):
                icon = resources.icon(icon)
            new_btn.setIcon(icon)
        else:
            new_btn.setText(name)
        new_btn.setObjectName(name.replace(' ', '').lower() or description)
        new_btn.setStatusTip(status or name)
        new_btn.setToolTip(tooltip or name)
        new_btn.setProperty('name', name)
        new_btn.setProperty('description', description or tooltip or status or '')
        new_btn.setProperty('instructions', instructions or '')
        new_btn.setProperty('tooltip_help', {'title': name, 'description': description})
        new_btn.setStyleSheet('qproperty-iconSize: 20px 20px;')
        new_btn.setIconSize(QSize(20, 20))
        new_btn.setWhatsThis(whats_this or tooltip or name)
        new_btn.setContextMenuPolicy(Qt.CustomContextMenu)
        new_btn.installEventFilter(self)
        if widgets:
            valid_widgets = list()
            for widget in widgets:
                if hasattr(self, widget):
                    valid_widgets.append(getattr(self, widget))
            if valid_widgets:
                self._widgets[new_btn.objectName()] = valid_widgets
                new_btn.has_menu = True

        if fn:
            info_message = info_message or 'Command "{}" executed successfully!'.format(name)
            error_message = error_message or 'Error while executing command "{}"! Check log for more info.'.format(name)
            new_btn.clicked.connect(partial(self._on_execute_command, name, fn, info_message, error_message))
        new_btn.customContextMenuRequested.connect(self._on_show_context_menu)

        if settings:
            settings_icon = 'menu_dots' if settings and len(settings) > 1 else 'settings'
            buttons_widget = QWidget()
            buttons_widget.setStyleSheet('QPushButton::menu-indicator{ image: none; };')
            buttons_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
            buttons_widget.setLayout(buttons_layout)

            if len(settings) > 1:
                options_name = '{} Options'.format(name)
                options_description = 'Show {} Menu'.format(options_name)
                options_btn = self._create_button(
                    icon=settings_icon, fn=settings_fn, name=options_name, description=options_description)
                options_menu = QMenu(self)
                options_menu.installEventFilter(self)
                options_btn.setMenu(options_menu)
                for setting in settings:
                    option_name = setting.get('name', None)
                    if not option_name:
                        continue
                    option_fn = setting.get('fn', None)
                    if not option_fn:
                        continue
                    option_description = setting.get('description', '')
                    option_instructions = setting.get('instructions', '')
                    option_icon = resources.icon(setting.get('icon', 'tpRigToolkit'))
                    option_action = QAction(option_icon, option_name, self)
                    option_action.setProperty('description', option_description or '')
                    option_action.setProperty('instructions', option_instructions or '')
                    option_info_message = info_message or 'Command "{}" executed successfully!'.format(name)
                    option_error_message = error_message or 'Error while executing command "{}"! ' \
                                                            'Check log for more info.'.format(name)
                    option_action.triggered.connect(
                        partial(self._on_execute_command, option_name,
                                option_fn, option_info_message, option_error_message))

                    options_menu.addAction(option_action)
            else:
                setting = settings[0]
                option_description = setting.get('description', '')
                option_name = setting.get('name', None)
                option_fn = setting.get('fn', None)
                options_btn = self._create_button(
                    icon=settings_icon, fn=option_fn, name=option_name, description=option_description)

            new_btn.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)
            new_btn.setStyleSheet(
                new_btn.styleSheet() + ';border-top-right-radius: 0px; border-bottom-right-radius: 0px; '
                                       'border-right: 0px;')
            new_btn.setMaximumWidth(28)

            options_btn.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)
            options_btn.setMaximumWidth(18)
            options_btn.setStyleSheet(
                options_btn.styleSheet() + ';border-top-left-radius: 0px; border-bottom-left-radius: 0px; '
                                           'qproperty-iconSize: 12px 12px;')
            buttons_layout.addWidget(new_btn)
            buttons_layout.addWidget(options_btn)

            self._buttons.append(new_btn)

            return buttons_widget

        self._buttons.append(new_btn)

        return new_btn

    def _set_info(self, name, description, instructions):
        self._info_widget.name = name
        self._info_widget.description = description
        self._info_widget.instructions = instructions
        self._info_label.setText(name)

    def _on_show_whats_this(self, whats_this_text):
        pos = QCursor.pos()
        if not whats_this_text:
            return
        QWhatsThis.showText(pos, whats_this_text, self)

    def _on_show_context_menu(self):
        sender = self.sender()
        if not sender:
            return

        sender_name = sender.objectName()
        if sender_name not in self._widgets:
            return

        menu = CommandBalloon(widgets_parent=self, parent=sender)
        widgets = self._widgets[sender_name]
        for widget in widgets:
            menu.add_widget(widget)
        menu.closed.connect(partial(self._on_closed_menu, sender))
        sender.primary()
        menu.show()

        pos = menu.mapToGlobal(QPoint(menu.width() / 2, 0))
        final_pos = sender.mapFromGlobal(pos)
        menu.move(
            menu.pos().x() - final_pos.x() + sender.width() / 2, menu.pos().y() - final_pos.y() + sender.height() - 5)

        return menu

    def _on_closed_menu(self, command_button):
        command_button.default()

    def _on_toggle_info(self, flag):
        self._info_frame.setVisible(not flag)
        self._info_widget.setVisible(flag)

    def _on_execute_command(self, name, fn, info_message=None, error_message=None):
        library.Command.startCommand.emit(name)
        start_time = time.time()
        valid_fn = fn()
        end_time = time.time() - start_time
        if valid_fn and info_message:
            info_message += ' (%0.3f secs)' % end_time
            self.emitInfo.emit(info_message)
        if not valid_fn and error_message:
            self.emitError.emit(error_message)
        library.Command.endCommand.emit(bool(valid_fn))

        return valid_fn


class CommandButton(buttons.BaseButton, object):
    def __init__(self, text='', icon=None, parent=None):
        super(CommandButton, self).__init__(text=text, icon=icon, parent=parent)

        self._has_menu = False

    @property
    def has_menu(self):
        return self._has_menu

    @has_menu.setter
    def has_menu(self, flag):
        self._has_menu = flag

    def paintEvent(self, event):
        super(CommandButton, self).paintEvent(event)
        if self.has_menu:
            painter = QPainter()
            painter.begin(self)
            brush = QBrush(QColor(self.theme().accent_color))
            painter.setRenderHint(painter.Antialiasing)
            painter.setBrush(brush)
            painter.setPen(Qt.NoPen)
            w = self.rect().width() - 1
            h = self.rect().height() - 1
            polygon = QPolygon()
            polygon.append(QPoint(w - 1, h - 8))
            polygon.append(QPoint(w - 8, h - 1))
            polygon.append(QPoint(w - 1, h - 1))
            painter.drawPolygon(polygon)


class CommandBalloon(balloon.BalloonDialog, object):
    def __init__(self, widgets_parent, modal=False, parent=None):
        super(CommandBalloon, self).__init__(modal=modal, parent=parent)

        self._widgets = list()
        self._widgets_parent = widgets_parent

        self.main_layout = layouts.VerticalLayout(spacing=2, margins=(2, 2, 2, 2))
        self.setLayout(self.main_layout)

    def add_widget(self, widget):
        self.main_layout.addWidget(widget)
        widget.setVisible(True)
        self._widgets.append(widget)

    def closeEvent(self, event):
        for widget in self._widgets:
            widget.setVisible(False)
            widget.setParent(self._widgets_parent)
        super(CommandBalloon, self).closeEvent(event)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains implementation for info command widget
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, Property
from Qt.QtWidgets import QSizePolicy, QWidget, QFrame, QPlainTextEdit, QTextEdit

from tpDcc.managers import resources
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import base
from tpDcc.libs.qt.widgets import layouts, message, expandables, dividers


@theme.mixin
class InfoMessage(base.BaseWidget, object):
    def __init__(self, name='', description='', instructions='', parent=None):
        self._name = ''
        self._description = ''
        self._instructions = instructions

        super(InfoMessage, self).__init__(parent)

        self.setAttribute(Qt.WA_StyledBackground)
        self.theme_type = message.MessageTypes.INFO
        self.style().polish(self)
        self.name = name
        self.description = description
        self.instructions = instructions

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    def _get_name(self):
        return self._name

    def _set_name(self, name):
        self._name = str(name)
        self._expandable_frame.set_title(self._name)

    def _get_description(self):
        return self._description

    def _set_description(self, text):
        self._description = str(text)
        self._description_text.setPlainText(self._description)

    def _get_instructions(self):
        return self._instructions

    def _set_instructions(self, instructions):
        self._instructions = str(instructions)
        self._instructions_text.clear()
        self._instructions_text.insertHtml(instructions)
        self._instructions_widget.setVisible(bool(instructions))

    name = Property(str, _get_name, _set_name)
    description = Property(str, _get_description, _set_description)
    instructions = Property(str, _get_instructions, _set_instructions)

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def ui(self):
        super(InfoMessage, self).ui()

        self.setMaximumHeight(150)

        info_icon = resources.icon('info')
        self._expandable_frame = expandables.ExpandableFrame(icon=info_icon, parent=self)
        self._expandable_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)

        expandable_layout = layouts.HorizontalLayout(margins=(2, 2, 2, 2))

        texts_layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._description_text = QPlainTextEdit(parent=self)
        self._description_text.setReadOnly(True)
        self._description_text.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        self._description_text.setFocusPolicy(Qt.NoFocus)
        self._description_text.setFrameShape(QFrame.NoFrame)
        self._instructions_widget = QWidget()
        instructions_layout = layouts.VerticalLayout(spacing=2, margins=(0, 0, 0, 0))
        self._instructions_widget.setLayout(instructions_layout)
        self._instructions_text = QTextEdit(parent=self)
        self._instructions_text.setReadOnly(True)
        self._instructions_text.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        self._instructions_text.setFocusPolicy(Qt.NoFocus)
        self._instructions_text.setFrameShape(QFrame.NoFrame)
        self._instructions_widget.setVisible(False)
        # self._instructions_text.insertHtml("<ul><li>text 1</li><li>text 2</li><li>text 3</li></ul> <br />")
        instructions_layout.addWidget(dividers.Divider('Instructions'))
        instructions_layout.addWidget(self._instructions_text)
        texts_layout.addWidget(self._description_text)
        texts_layout.addWidget(self._instructions_widget)

        content_layout = layouts.VerticalLayout()
        content_layout.addLayout(texts_layout)
        expandable_layout.addLayout(content_layout)

        self._expandable_frame.addLayout(expandable_layout)

        self.main_layout.addWidget(self._expandable_frame)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains basic library implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import QObject, Signal


class _CommandSignals(QObject, object):

    command = None

    startCommand = Signal(str)
    progressCommand = Signal(int, str)
    endCommand = Signal(bool)


Command = _CommandSignals()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library with joint functions
"""

from __future__ import print_function, division, absolute_import

import maya.cmds

from tpDcc import dcc
from tpDcc.libs.python import python

from tpDcc.dccs.maya.core import decorators, joint as joint_utils, curve as curve_utils


def get_valid_joints(joints=None):

    valid_joints = list()
    joints = joints or dcc.selected_nodes_of_type(node_type='joint', full_path=False)
    joints = python.force_list(joints)
    for joint in joints:
        if not dcc.object_type(joint) == 'joint':
            continue
        valid_joints.append(joint)
    if not valid_joints:
        valid_joints = dcc.list_nodes(node_type='joint')

    return valid_joints


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def start_joint_tool():
    """
    Initializes joint tool
    """

    out_dict = {'success': False, 'result': list()}

    try:
        dcc.start_joint_tool()
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to open joint tool: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def create_new_joint_on_center(transforms=None):
    """
    Creates a new joint in the center of the selected transforms
    """

    out_dict = {'success': False, 'result': None}

    try:
        selection = maya.cmds.ls(sl=True)
        if transforms:
            result = joint_utils.create_joint_on_center()
        else:
            if '.[' in selection:
                result = joint_utils.create_joints_on_selected_components_center()
            else:
                result = joint_utils.create_joint_on_center()
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to open joint tool: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def create_joints_on_selected_components_center():
    """
    Creates new joints on the selected components center
    """

    out_dict = {'success': False, 'result': None}

    try:

        new_joints = joint_utils.create_joints_on_selected_components_center()
        out_dict['result'] = new_joints
    except Exception as exc:
        out_dict['success'] = False
        out_dict['msg'] = 'Was not possible to create joints on selected components center: {}'.format(exc)
        return out_dict

    if not new_joints:
        out_dict['success'] = False
        out_dict['msg'] = 'No components to create joints on'
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def insert_joints(joints=None, num_joints=1):
    """
    Inserts new joints between selected joint and its direct child
    :param joints: int
    :param num_joints: int
    """

    out_dict = {'success': False, 'result': None}

    joints = joints or dcc.selected_nodes_of_type(node_type='joint')
    joints = python.force_list(joints)
    if not joints:
        out_dict['msg'] = 'No joints to insert joints into found.'
        return out_dict

    try:

        result = joint_utils.insert_joints(joints=joints, joint_count=num_joints)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to insert joints: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def create_joints_on_curve(curve=None, num_joints=1):

    out_dict = {'success': False, 'result': list()}

    valid_curves = list()
    curves = curve or dcc.selected_nodes_of_type(node_type='transform')
    curves = python.force_list(curves)
    for curve in curves:
        if not curve_utils.is_a_curve(curve):
            continue
        valid_curves.append(curve)
    if not valid_curves:
        out_dict['msg'] = 'No curves to create joints on found.'
        return out_dict

    try:
        for curve in valid_curves:
            new_joint = joint_utils.create_oriented_joints_along_curve(curve, num_joints)
            out_dict['result'].append(new_joint)
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to create joints on curve: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def snap_joints_to_curve(joints=None, curve=None, num_joints=1):

    out_dict = {'success': False, 'result': dict()}

    valid_joints = list()
    joints = joints or dcc.selected_nodes_of_type(node_type='joint')
    joints = python.force_list(joints)
    for joint in joints:
        if not dcc.object_type(joint) == 'joint':
            continue
        valid_joints.append(joint)
    if not valid_joints:
        out_dict['msg'] = 'No joints to snap to curve found.'
        return out_dict

    valid_curves = list()
    curves = curve or dcc.selected_nodes_of_type(node_type='transform')
    curves = python.force_list(curves)
    for curve in curves:
        if not curve_utils.is_a_curve(curve):
            continue
        valid_curves.append(curve)

    try:
        for curve in valid_curves:
            new_joints = curve_utils.snap_joints_to_curve(valid_joints, curve=curve, count=num_joints)
            out_dict['result'][curve] = new_joints
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to snap joints to curve: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


def check_joint_labels(joints=None):

    out_dict = {'success': False, 'result': False}

    valid_joints = list()
    joints = joints or dcc.selected_nodes_of_type(node_type='joint')
    joints = python.force_list(joints)
    for joint in joints:
        if not dcc.object_type(joint) == 'joint':
            continue
        valid_joints.append(joint)
    if not valid_joints:
        out_dict['msg'] = 'No joints to check found.'
        return out_dict

    try:
        result = joint_utils.check_joint_labels(valid_joints)
        out_dict['result'] = result
    except Exception as exc:
        out_dict['msg'] = 'Was not possible to check joints label: {}'.format(exc)
        return out_dict

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def toggle_local_rotation_axis(joints=None):
    """
     Toggles the visibility of all (if the user has nothing selected) or all joints local rotation axis
     """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = list()
    joints = joints or dcc.selected_nodes_of_type(node_type='joint')
    joints = python.force_list(joints)
    for joint in joints:
        if not dcc.object_type(joint) == 'joint':
            continue
        valid_joints.append(joint)
    if not valid_joints:
        valid_joints = None

    dcc.set_joint_local_rotation_axis_visibility(flag=None, joints_to_apply=valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def toggle_all_local_rotation_axis(flag=None):
    """
    Toggles the visibility of all joints local rotation axis
    """

    out_dict = {'success': False, 'result': dict()}

    dcc.set_joint_local_rotation_axis_visibility(flag=flag, joints_to_apply=None)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def toggle_selected_local_rotation_axis(flag=None):
    """
    Toggles the visibility of selected joints local rotation axis
    """

    out_dict = {'success': False, 'result': dict()}

    joints = dcc.selected_nodes_of_type(node_type='joint')
    if not joints:
        out_dict['msg'] = 'No joints to toggle LRA of found.'
        return out_dict

    dcc.set_joint_local_rotation_axis_visibility(flag=flag, joints_to_apply=joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def toggle_joints_xray():
    out_dict = {'success': False, 'result': dict()}

    dcc.toggle_xray_joints()

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def set_joints_xray(flag=None):
    out_dict = {'success': False, 'result': dict()}

    if flag is None:
        dcc.toggle_xray_joints()
    else:
        current_panel = maya.cmds.getPanel(withFocus=True)
        maya.cmds.modelEditor(current_panel, edit=True, jointXray=flag)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def set_joints_display_size(value=None):
    out_dict = {'success': False, 'result': dict()}

    if value is None:
        value = dcc.get_joint_display_size()

    dcc.set_joint_display_size(value)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def select_hierarchy():
    out_dict = {'success': False, 'result': dict()}

    dcc.select_hierarchy()

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def orient_joints(joints=None, force_orient_attributes=True):
    """
    Orients all joints that have OrientJointAttributes added
    :param joints: list(str) or None
    :param force_orient_attributes: bool, Whether to force the orientation through OrientJointAttributes or not
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints(joints)
    if not valid_joints:
        valid_joints = dcc.list_nodes(node_type='joint')
    if not valid_joints:
        out_dict['msg'] = 'No joints to orient found.'
        return out_dict

    dcc.orient_joints(joints_to_orient=valid_joints, force_orient_attributes=force_orient_attributes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def orient_all_joints(force_orient_attributes=True):
    """
    Orients all joints
    :param force_orient_attributes: bool, Whether to force the orientation through OrientJointAttributes or not
    """

    out_dict = {'success': False, 'result': dict()}

    all_joints = dcc.list_nodes(node_type='joint', full_path=False)
    if not all_joints:
        out_dict['msg'] = 'No joints to orient found.'
        return out_dict

    dcc.orient_joints(joints_to_orient=all_joints, force_orient_attributes=force_orient_attributes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def orient_selected_joints(force_orient_attributes=True):
    """
    Orient selected joints
    :param force_orient_attributes: bool, Whether to force the orientation through OrientJointAttributes or not
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints()
    if not valid_joints:
        out_dict['msg'] = 'No joints to orient found.'
        return out_dict

    dcc.orient_joints(joints_to_orient=valid_joints, force_orient_attributes=force_orient_attributes)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def add_orient_data(joints=None):
    """
    Add orient data to joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints(joints)
    if not valid_joints:
        valid_joints = dcc.list_nodes(node_type='joint')
    if not valid_joints:
        out_dict['msg'] = 'No joints to add orient data to found.'
        return out_dict

    joint_utils.OrientJointAttributes.add_orient_attributes(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def add_orient_data_all_joints():
    """
    Add orient data to all joints in current scene
    """

    out_dict = {'success': False, 'result': dict()}

    all_joints = dcc.list_nodes(node_type='joint', full_path=False)
    if not all_joints:
        out_dict['msg'] = 'No joints to add orient data to found.'
        return out_dict

    joint_utils.OrientJointAttributes.add_orient_attributes(all_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def add_orient_data_selected_joints():
    """
    Add orient data to selected joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints()
    if not valid_joints:
        out_dict['msg'] = 'No joints to add orient data to found.'
        return out_dict

    joint_utils.OrientJointAttributes.add_orient_attributes(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def clean_orient_data(joints=None):
    """
    Clean orient data from joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints(joints)
    if not valid_joints:
        valid_joints = dcc.list_nodes(node_type='joint')
    if not valid_joints:
        out_dict['msg'] = 'No joints to remove orient data from found.'
        return out_dict

    joint_utils.OrientJointAttributes.remove_orient_attributes(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def clean_orient_data_all_joints():
    """
    Clean orient data from all joints in current scene
    """

    out_dict = {'success': False, 'result': dict()}

    all_joints = dcc.list_nodes(node_type='joint', full_path=False)
    if not all_joints:
        out_dict['msg'] = 'No joints to remove orient data from found.'
        return out_dict

    joint_utils.OrientJointAttributes.remove_orient_attributes(all_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def clean_orient_data_selected_joints():
    """
    Cleans orient data from selected joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints()
    if not valid_joints:
        out_dict['msg'] = 'No joints to remove orient data from found.'
        return out_dict

    joint_utils.OrientJointAttributes.remove_orient_attributes(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def zero_joint_orient(joints=None):
    """
    Zeroes out the data of joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints(joints)
    if not valid_joints:
        valid_joints = dcc.list_nodes(node_type='joint')
    if not valid_joints:
        out_dict['msg'] = 'No joints to zero out orient of found.'
        return out_dict

    joint_utils.OrientJointAttributes.zero_orient_joint(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def zero_joint_orient_all_joints():
    """
    Zeroes out the orient of all joints in current scene
    """

    out_dict = {'success': False, 'result': dict()}

    all_joints = dcc.list_nodes(node_type='joint', full_path=False)
    if not all_joints:
        out_dict['msg'] = 'No joints to remove orient data from found.'
        return out_dict

    joint_utils.OrientJointAttributes.zero_orient_joint(all_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def zero_joint_orient_selected_joints():
    """
    Zeroes out the orient of selected joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints()
    if not valid_joints:
        out_dict['msg'] = 'No joints to remove orient data from found.'
        return out_dict

    joint_utils.OrientJointAttributes.zero_orient_joint(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def mirror_joints(joints=None):
    """
    Mirror joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints(joints)
    if not valid_joints:
        valid_joints = dcc.list_nodes(node_type='joint')
    if not valid_joints:
        out_dict['msg'] = 'No joints to mirror found.'
        return out_dict

    dcc.mirror_transform(valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def mirror_all_joints():
    """
    Mirror all joints in current scene
    """

    out_dict = {'success': False, 'result': dict()}

    all_joints = dcc.list_nodes(node_type='joint', full_path=False)
    if not all_joints:
        out_dict['msg'] = 'No joints to mirror found.'
        return out_dict

    dcc.mirror_transform(all_joints, create_if_missing=True)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def mirror_selected_joints():
    """
    Mirror selected joints
    """

    out_dict = {'success': False, 'result': dict()}

    valid_joints = get_valid_joints()
    if not valid_joints:
        out_dict['msg'] = 'No joints to mirror found.'
        return out_dict

    dcc.mirror_transform(create_if_missing=True, transforms=valid_joints)

    out_dict['success'] = True

    return out_dict


@decorators.undo
@decorators.repeat_static_command(__name__, skip_arguments=True)
def mirror_hierarchy_joints():
    """
    Mirror hierarchy joints
    """

    out_dict = {'success': False, 'result': dict()}

    all_joints = list()

    valid_joints = get_valid_joints()
    if not valid_joints:
        out_dict['msg'] = 'No joints to mirror found.'
        return out_dict

    for joint in valid_joints:
        all_joints.append(joint)
        children = dcc.list_children(joint, full_path=False, children_type='transform')
        found = list()
        if children:
            for child in children:
                if dcc.node_type(child).find('Constraint') > -1:
                    continue
                found.append(child)
        all_joints.extend(found)

    all_joints = list(set(all_joints))

    dcc.mirror_transform(transforms=all_joints, create_if_missing=True)

    out_dict['success'] = True

    return out_dict
